from nmigen import Signal, Module, Elaboratable
from nmigen.compat.sim import run_simulation

# test for https://github.com/nmigen/nmigen/issues/344


class MyModule(Elaboratable):
    def __init__(self):
        self.a = Signal()

    def elaborate(self, platform):
        m = Module()
        m.d.sync += self.a.eq(~self.a)
        return m


def test1():
    dut = MyModule()

    def generator():
        for _i in range(10):
            print((yield dut.a))
            yield

    run_simulation(dut, generator(),
                   vcd_name="test_run_simulation_bug.vcd")


if __name__ == '__main__':
    test1()



# Copyright (c) 2014 - 2019 The Regents of the University of
# California (Regents). All Rights Reserved.  Redistribution and use in
# source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
#    * Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      two paragraphs of disclaimer.
#    * Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      two paragraphs of disclaimer in the documentation and/or other materials
#      provided with the distribution.
#    * Neither the name of the Regents nor the names of its contributors
#      may be used to endorse or promote products derived from this
#      software without specific prior written permission.
# IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
# SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
# ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
# REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
# ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS". REGENTS HAS NO OBLIGATION
# TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
# MODIFICATIONS.

from nmigen.compat.fhdl.specials import Memory
from nmigen import Module, Signal, Mux, Elaboratable
from nmigen.utils import bits_for
from nmigen.cli import main
from nmigen.lib.fifo import FIFOInterface

# translated from https://github.com/freechipsproject/chisel3/blob/a4a29e29c3f1eed18f851dcf10bdc845571dfcb6/src/main/scala/chisel3/util/Decoupled.scala#L185   # noqa


class Queue(FIFOInterface, Elaboratable):
    def __init__(self, width, depth, fwft=True, pipe=False):
        """ Queue (FIFO) with pipe mode and first-write fall-through capability

            * :width: width of Queue data in/out
            * :depth: queue depth.  NOTE: may be set to 0 (this is ok)
            * :fwft : first-write, fall-through mode (Chisel Queue "flow" mode)
            * :pipe : pipe mode.  NOTE: this mode can cause unanticipated
                      problems.  when read is enabled, so is w_rdy.
                      therefore if read is enabled, the data ABSOLUTELY MUST
                      be read.

            fwft mode = True basically means that the data may be transferred
            combinatorially from input to output.

            Attributes:
            * level: available free space (number of unread entries)

            w_data  = enq_data, w_rdy  = enq_ready, w_en = enq_valid
            r_data = deq_data, r_en = deq_ready, r_rdy = deq_valid
        """
        FIFOInterface.__init__(self, width=width, depth=depth, fwft=fwft)
        self.pipe = pipe
        self.depth = depth
        self.level = Signal(bits_for(depth))

    def elaborate(self, platform):
        m = Module()

        # set up an SRAM.  XXX bug in Memory: cannot create SRAM of depth 1
        ram = Memory(self.width, self.depth if self.depth > 1 else 2)
        m.submodules.ram = ram
        m.submodules.ram_read = ram_read = ram.read_port(domain="comb")
        m.submodules.ram_write = ram_write = ram.write_port()

        # convenience names, for people familiar with ready/valid terminology
        # "p" stands for "previous stage", "n" stands for "next stage"
        # for people familiar with the chisel Decoupled library:
        # enq is "enqueue" (data in, aka "prev stage"),
        # deq is "dequeue" (data out, aka "next stage")
        p_ready_o = self.w_rdy
        p_valid_i = self.w_en
        enq_data = self.w_data # aka p_data_i

        n_valid_o = self.r_rdy
        n_ready_i = self.r_en
        deq_data = self.r_data # aka n_data_o

        # intermediaries
        ptr_width = bits_for(self.depth - 1) if self.depth > 1 else 0
        enq_ptr = Signal(ptr_width) # cyclic pointer to "insert" point (wrport)
        deq_ptr = Signal(ptr_width) # cyclic pointer to "remove" point (rdport)
        maybe_full = Signal() # not reset_less (set by sync)

        # temporaries
        do_enq = Signal(reset_less=True)
        do_deq = Signal(reset_less=True)
        ptr_diff = Signal(ptr_width)
        ptr_match = Signal(reset_less=True)
        empty = Signal(reset_less=True)
        full = Signal(reset_less=True)
        enq_max = Signal(reset_less=True)
        deq_max = Signal(reset_less=True)

        m.d.comb += [ptr_match.eq(enq_ptr == deq_ptr), # read-ptr = write-ptr
                     ptr_diff.eq(enq_ptr - deq_ptr),
                     enq_max.eq(enq_ptr == self.depth - 1),
                     deq_max.eq(deq_ptr == self.depth - 1),
                     empty.eq(ptr_match & ~maybe_full),
                     full.eq(ptr_match & maybe_full),
                     do_enq.eq(p_ready_o & p_valid_i), # write conditions ok
                     do_deq.eq(n_ready_i & n_valid_o), # read conditions ok

                     # set r_rdy and w_rdy (NOTE: see pipe mode below)
                     n_valid_o.eq(~empty), # cannot read if empty!
                     p_ready_o.eq(~full),  # cannot write if full!

                     # set up memory and connect to input and output
                     ram_write.addr.eq(enq_ptr),
                     ram_write.data.eq(enq_data),
                     ram_write.en.eq(do_enq),
                     ram_read.addr.eq(deq_ptr),
                     deq_data.eq(ram_read.data) # NOTE: overridden in fwft mode
                    ]

        # under write conditions, SRAM write-pointer moves on next clock
        with m.If(do_enq):
            m.d.sync += enq_ptr.eq(Mux(enq_max, 0, enq_ptr+1))

        # under read conditions, SRAM read-pointer moves on next clock
        with m.If(do_deq):
            m.d.sync += deq_ptr.eq(Mux(deq_max, 0, deq_ptr+1))

        # if read-but-not-write or write-but-not-read, maybe_full set
        with m.If(do_enq != do_deq):
            m.d.sync += maybe_full.eq(do_enq)

        # first-word fall-through: same as "flow" parameter in Chisel3 Queue
        # basically instead of relying on the Memory characteristics (which
        # in FPGAs do not have write-through), then when the queue is empty
        # take the output directly from the input, i.e. *bypass* the SRAM.
        # this done combinatorially to give the exact same characteristics
        # as Memory "write-through"... without relying on a changing API
        if self.fwft:
            with m.If(p_valid_i):
                m.d.comb += n_valid_o.eq(1)
            with m.If(empty):
                m.d.comb += deq_data.eq(enq_data)
                m.d.comb += do_deq.eq(0)
                with m.If(n_ready_i):
                    m.d.comb += do_enq.eq(0)

        # pipe mode: if next stage says it's ready (r_rdy), w_en
        #            *must* declare the input ready (w_rdy).
        if self.pipe:
            with m.If(n_ready_i):
                m.d.comb += p_ready_o.eq(1)

        # set the count (available free space), optimise on power-of-two
        if self.depth == 1 << ptr_width:  # is depth a power of 2
            m.d.comb += self.level.eq(
                Mux(maybe_full & ptr_match, self.depth, 0) | ptr_diff)
        else:
            m.d.comb += self.level.eq(Mux(ptr_match,
                                          Mux(maybe_full, self.depth, 0),
                                          Mux(deq_ptr > enq_ptr,
                                              self.depth + ptr_diff,
                                              ptr_diff)))

        return m


if __name__ == "__main__":
    reg_stage = Queue(1, 1, pipe=True)
    break_ready_chain_stage = Queue(1, 1, pipe=True, fwft=True)
    m = Module()
    ports = []

    def queue_ports(queue, name_prefix):
        retval = []
        for name in ["level",
                     "r_data",
                     "r_rdy",
                     "w_rdy"]:
            port = getattr(queue, name)
            signal = Signal(port.shape(), name=name_prefix+name)
            m.d.comb += signal.eq(port)
            retval.append(signal)
        for name in ["r_en",
                     "w_data",
                     "w_en"]:
            port = getattr(queue, name)
            signal = Signal(port.shape(), name=name_prefix+name)
            m.d.comb += port.eq(signal)
            retval.append(signal)
        return retval

    m.submodules.reg_stage = reg_stage
    ports += queue_ports(reg_stage, "reg_stage_")
    m.submodules.break_ready_chain_stage = break_ready_chain_stage
    ports += queue_ports(break_ready_chain_stage, "break_ready_chain_stage_")
    main(m, ports=ports)



"""Run-time selection of simulator engine

Usage::

    from nmutil.sim_tmp_alternative import Simulator

Then, use :py:class:`Simulator` as usual.

This should be backwards compatible to old developer versions of nMigen.

To use cxxsim, export ``NMIGEN_SIM_MODE=cxxsim`` from the shell.
Be sure to check out the ``cxxsim`` branch of nMigen, and update yosys
to the latest commit as well.

To use pysim, just keep ``NMIGEN_SIM_MODE`` unset.
Alternatively, export ``NMIGEN_SIM_MODE=pysim``.

Example::

    $ export NMIGEN_SIM_MODE=...  # pysim or cxxsim, default is pysim
    $ python ...

or, even::

    $ NMIGEN_SIM_MODE=...  python ...
"""

import os

try:
    from nmigen.sim import (Simulator as RealSimulator, Delay, Settle, Tick,
                            Passive)
    detected_new_api = True
except ImportError:
    detected_new_api = False
    try:
        from nmigen.sim.pysim import (Simulator as RealSimulator,
                                      Delay, Settle, Tick, Passive)
    except ImportError:
        from nmigen.back.pysim import (Simulator as RealSimulator,
                                       Delay, Settle, Tick, Passive)

nmigen_sim_environ_variable = os.environ.get("NMIGEN_SIM_MODE") \
                              or "pysim"
"""Detected run-time engine from environment"""


def Simulator(*args, **kwargs):
    """Wrapper that allows run-time selection of simulator engine"""
    if detected_new_api:
        kwargs['engine'] = nmigen_sim_environ_variable
    return RealSimulator(*args, **kwargs)


def is_engine_cxxsim():
    """Returns ``True`` if the selected engine is cxxsim"""
    return detected_new_api and nmigen_sim_environ_variable == "cxxsim"


def is_engine_pysim():
    """Returns ``True`` if the selected engine is pysim"""
    return not detected_new_api or nmigen_sim_environ_variable == "pysim"


nmigen_sim_top_module = "top." if is_engine_pysim() else ""
"""Work-around for cxxsim not defining the top-level module"""



""" Combinatorial Multi-input and Multi-output multiplexer blocks
    conforming to Pipeline API

    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+

    Multi-input is complex because if any one input is ready, the output
    can be ready, and the decision comes from a separate module.

    Multi-output is simple (pretty much identical to UnbufferedPipeline),
    and the selection is just a mux.  The only proviso (difference) being:
    the outputs not being selected have to have their ready_o signals
    DEASSERTED.

    https://bugs.libre-soc.org/show_bug.cgi?id=538
"""

from math import log
from nmigen import Signal, Cat, Const, Mux, Module, Array, Elaboratable
from nmigen.cli import verilog, rtlil
from nmigen.lib.coding import PriorityEncoder
from nmigen.hdl.rec import Record, Layout
from nmutil.stageapi import _spec

from collections.abc import Sequence

from nmutil.nmoperator import eq
from nmutil.iocontrol import NextControl, PrevControl


class MultiInControlBase(Elaboratable):
    """ Common functions for Pipeline API
    """
    def __init__(self, in_multi=None, p_len=1, maskwid=0, routemask=False):
        """ Multi-input Control class.  Conforms to same API as ControlBase...
            mostly.  has additional indices to the *multiple* input stages

            * p: contains ready/valid to the previous stages PLURAL
            * n: contains ready/valid to the next stage

            User must also:
            * add data_i members to PrevControl and
            * add data_o member  to NextControl
        """
        self.routemask = routemask
        # set up input and output IO ACK (prev/next ready/valid)
        print ("multi_in", self, maskwid, p_len, routemask)
        p = []
        for i in range(p_len):
            p.append(PrevControl(in_multi, maskwid=maskwid))
        self.p = Array(p)
        if routemask:
            nmaskwid = maskwid # straight route mask mode
        else:
            nmaskwid = maskwid * p_len # fan-in mode
        self.n = NextControl(maskwid=nmaskwid) # masks fan in (Cat)

    def connect_to_next(self, nxt, p_idx=0):
        """ helper function to connect to the next stage data/valid/ready.
        """
        return self.n.connect_to_next(nxt.p[p_idx])

    def _connect_in(self, prev, idx=0, prev_idx=None):
        """ helper function to connect stage to an input source.  do not
            use to connect stage-to-stage!
        """
        if prev_idx is None:
            return self.p[idx]._connect_in(prev.p)
        return self.p[idx]._connect_in(prev.p[prev_idx])

    def _connect_out(self, nxt):
        """ helper function to connect stage to an output source.  do not
            use to connect stage-to-stage!
        """
        if nxt_idx is None:
            return self.n._connect_out(nxt.n)
        return self.n._connect_out(nxt.n)

    def set_input(self, i, idx=0):
        """ helper function to set the input data
        """
        return eq(self.p[idx].data_i, i)

    def elaborate(self, platform):
        m = Module()
        for i, p in enumerate(self.p):
            setattr(m.submodules, "p%d" % i, p)
        m.submodules.n = self.n
        return m

    def __iter__(self):
        for p in self.p:
            yield from p
        yield from self.n

    def ports(self):
        return list(self)


class MultiOutControlBase(Elaboratable):
    """ Common functions for Pipeline API
    """
    def __init__(self, n_len=1, in_multi=None, maskwid=0, routemask=False):
        """ Multi-output Control class.  Conforms to same API as ControlBase...
            mostly.  has additional indices to the multiple *output* stages
            [MultiInControlBase has multiple *input* stages]

            * p: contains ready/valid to the previou stage
            * n: contains ready/valid to the next stages PLURAL

            User must also:
            * add data_i member to PrevControl and
            * add data_o members to NextControl
        """

        if routemask:
            nmaskwid = maskwid # straight route mask mode
        else:
            nmaskwid = maskwid * n_len # fan-out mode

        # set up input and output IO ACK (prev/next ready/valid)
        self.p = PrevControl(in_multi, maskwid=nmaskwid)
        n = []
        for i in range(n_len):
            n.append(NextControl(maskwid=maskwid))
        self.n = Array(n)

    def connect_to_next(self, nxt, n_idx=0):
        """ helper function to connect to the next stage data/valid/ready.
        """
        return self.n[n_idx].connect_to_next(nxt.p)

    def _connect_in(self, prev, idx=0):
        """ helper function to connect stage to an input source.  do not
            use to connect stage-to-stage!
        """
        return self.n[idx]._connect_in(prev.p)

    def _connect_out(self, nxt, idx=0, nxt_idx=None):
        """ helper function to connect stage to an output source.  do not
            use to connect stage-to-stage!
        """
        if nxt_idx is None:
            return self.n[idx]._connect_out(nxt.n)
        return self.n[idx]._connect_out(nxt.n[nxt_idx])

    def elaborate(self, platform):
        m = Module()
        m.submodules.p = self.p
        for i, n in enumerate(self.n):
            setattr(m.submodules, "n%d" % i, n)
        return m

    def set_input(self, i):
        """ helper function to set the input data
        """
        return eq(self.p.data_i, i)

    def __iter__(self):
        yield from self.p
        for n in self.n:
            yield from n

    def ports(self):
        return list(self)


class CombMultiOutPipeline(MultiOutControlBase):
    """ A multi-input Combinatorial block conforming to the Pipeline API

        Attributes:
        -----------
        p.data_i : stage input data (non-array).  shaped according to ispec
        n.data_o : stage output data array.       shaped according to ospec
    """

    def __init__(self, stage, n_len, n_mux, maskwid=0, routemask=False):
        MultiOutControlBase.__init__(self, n_len=n_len, maskwid=maskwid,
                                            routemask=routemask)
        self.stage = stage
        self.maskwid = maskwid
        self.routemask = routemask
        self.n_mux = n_mux

        # set up the input and output data
        self.p.data_i = _spec(stage.ispec, 'data_i') # input type
        for i in range(n_len):
            name = 'data_o_%d' % i
            self.n[i].data_o = _spec(stage.ospec, name) # output type

    def process(self, i):
        if hasattr(self.stage, "process"):
            return self.stage.process(i)
        return i

    def elaborate(self, platform):
        m = MultiOutControlBase.elaborate(self, platform)

        if hasattr(self.n_mux, "elaborate"): # TODO: identify submodule?
            m.submodules.n_mux = self.n_mux

        # need buffer register conforming to *input* spec
        r_data = _spec(self.stage.ispec, 'r_data') # input type
        if hasattr(self.stage, "setup"):
            self.stage.setup(m, r_data)

        # multiplexer id taken from n_mux
        muxid = self.n_mux.m_id
        print ("self.n_mux", self.n_mux)
        print ("self.n_mux.m_id", self.n_mux.m_id)

        self.n_mux.m_id.name = "m_id"

        # temporaries
        p_valid_i = Signal(reset_less=True)
        pv = Signal(reset_less=True)
        m.d.comb += p_valid_i.eq(self.p.valid_i_test)
        #m.d.comb += pv.eq(self.p.valid_i) #& self.n[muxid].ready_i)
        m.d.comb += pv.eq(self.p.valid_i & self.p.ready_o)

        # all outputs to next stages first initialised to zero (invalid)
        # the only output "active" is then selected by the muxid
        for i in range(len(self.n)):
            m.d.comb += self.n[i].valid_o.eq(0)
        if self.routemask:
            #with m.If(pv):
            m.d.comb += self.n[muxid].valid_o.eq(pv)
            m.d.comb += self.p.ready_o.eq(self.n[muxid].ready_i)
        else:
            data_valid = self.n[muxid].valid_o
            m.d.comb += self.p.ready_o.eq(~data_valid | self.n[muxid].ready_i)
            m.d.comb += data_valid.eq(p_valid_i | \
                                    (~self.n[muxid].ready_i & data_valid))


        # send data on
        #with m.If(pv):
        m.d.comb += eq(r_data, self.p.data_i)
        m.d.comb += eq(self.n[muxid].data_o, self.process(r_data))

        if self.maskwid:
            if self.routemask: # straight "routing" mode - treat like data
                m.d.comb += self.n[muxid].stop_o.eq(self.p.stop_i)
                with m.If(pv):
                    m.d.comb += self.n[muxid].mask_o.eq(self.p.mask_i)
            else:
                ml = [] # accumulate output masks
                ms = [] # accumulate output stops
                # fan-out mode.
                # conditionally fan-out mask bits, always fan-out stop bits
                for i in range(len(self.n)):
                    ml.append(self.n[i].mask_o)
                    ms.append(self.n[i].stop_o)
                m.d.comb += Cat(*ms).eq(self.p.stop_i)
                with m.If(pv):
                    m.d.comb += Cat(*ml).eq(self.p.mask_i)
        return m


class CombMultiInPipeline(MultiInControlBase):
    """ A multi-input Combinatorial block conforming to the Pipeline API

        Attributes:
        -----------
        p.data_i : StageInput, shaped according to ispec
            The pipeline input
        p.data_o : StageOutput, shaped according to ospec
            The pipeline output
        r_data : input_shape according to ispec
            A temporary (buffered) copy of a prior (valid) input.
            This is HELD if the output is not ready.  It is updated
            SYNCHRONOUSLY.
    """

    def __init__(self, stage, p_len, p_mux, maskwid=0, routemask=False):
        MultiInControlBase.__init__(self, p_len=p_len, maskwid=maskwid,
                                          routemask=routemask)
        self.stage = stage
        self.maskwid = maskwid
        self.p_mux = p_mux

        # set up the input and output data
        for i in range(p_len):
            name = 'data_i_%d' % i
            self.p[i].data_i = _spec(stage.ispec, name) # input type
        self.n.data_o = _spec(stage.ospec, 'data_o')

    def process(self, i):
        if hasattr(self.stage, "process"):
            return self.stage.process(i)
        return i

    def elaborate(self, platform):
        m = MultiInControlBase.elaborate(self, platform)

        m.submodules.p_mux = self.p_mux

        # need an array of buffer registers conforming to *input* spec
        r_data = []
        data_valid = []
        p_valid_i = []
        n_ready_in = []
        p_len = len(self.p)
        for i in range(p_len):
            name = 'r_%d' % i
            r = _spec(self.stage.ispec, name) # input type
            r_data.append(r)
            data_valid.append(Signal(name="data_valid", reset_less=True))
            p_valid_i.append(Signal(name="p_valid_i", reset_less=True))
            n_ready_in.append(Signal(name="n_ready_in", reset_less=True))
            if hasattr(self.stage, "setup"):
                print ("setup", self, self.stage, r)
                self.stage.setup(m, r)
        if len(r_data) > 1:
            r_data = Array(r_data)
            p_valid_i = Array(p_valid_i)
            n_ready_in = Array(n_ready_in)
            data_valid = Array(data_valid)

        nirn = Signal(reset_less=True)
        m.d.comb += nirn.eq(~self.n.ready_i)
        mid = self.p_mux.m_id
        print ("CombMuxIn mid", self, self.stage, self.routemask, mid, p_len)
        for i in range(p_len):
            m.d.comb += data_valid[i].eq(0)
            m.d.comb += n_ready_in[i].eq(1)
            m.d.comb += p_valid_i[i].eq(0)
            #m.d.comb += self.p[i].ready_o.eq(~data_valid[i] | self.n.ready_i)
            m.d.comb += self.p[i].ready_o.eq(0)
        p = self.p[mid]
        maskedout = Signal(reset_less=True)
        if hasattr(p, "mask_i"):
            m.d.comb += maskedout.eq(p.mask_i & ~p.stop_i)
        else:
            m.d.comb += maskedout.eq(1)
        m.d.comb += p_valid_i[mid].eq(maskedout & self.p_mux.active)
        m.d.comb += self.p[mid].ready_o.eq(~data_valid[mid] | self.n.ready_i)
        m.d.comb += n_ready_in[mid].eq(nirn & data_valid[mid])
        anyvalid = Signal(i, reset_less=True)
        av = []
        for i in range(p_len):
            av.append(data_valid[i])
        anyvalid = Cat(*av)
        m.d.comb += self.n.valid_o.eq(anyvalid.bool())
        m.d.comb += data_valid[mid].eq(p_valid_i[mid] | \
                                    (n_ready_in[mid] ))

        if self.routemask:
            # XXX hack - fixes loop
            m.d.comb += eq(self.n.stop_o, self.p[-1].stop_i)
            for i in range(p_len):
                p = self.p[i]
                vr = Signal(name="vr%d" % i, reset_less=True)
                maskedout = Signal(name="maskedout%d" % i, reset_less=True)
                if hasattr(p, "mask_i"):
                    m.d.comb += maskedout.eq(p.mask_i & ~p.stop_i)
                else:
                    m.d.comb += maskedout.eq(1)
                m.d.comb += vr.eq(maskedout.bool() & p.valid_i & p.ready_o)
                #m.d.comb += vr.eq(p.valid_i & p.ready_o)
                with m.If(vr):
                    m.d.comb += eq(self.n.mask_o, self.p[i].mask_i)
                    m.d.comb += eq(r_data[i], self.p[i].data_i)
        else:
            ml = [] # accumulate output masks
            ms = [] # accumulate output stops
            for i in range(p_len):
                vr = Signal(reset_less=True)
                p = self.p[i]
                vr = Signal(reset_less=True)
                maskedout = Signal(reset_less=True)
                if hasattr(p, "mask_i"):
                    m.d.comb += maskedout.eq(p.mask_i & ~p.stop_i)
                else:
                    m.d.comb += maskedout.eq(1)
                m.d.comb += vr.eq(maskedout.bool() & p.valid_i & p.ready_o)
                with m.If(vr):
                    m.d.comb += eq(r_data[i], self.p[i].data_i)
                if self.maskwid:
                    mlen = len(self.p[i].mask_i)
                    s = mlen*i
                    e = mlen*(i+1)
                    ml.append(Mux(vr, self.p[i].mask_i, Const(0, mlen)))
                    ms.append(self.p[i].stop_i)
            if self.maskwid:
                m.d.comb += self.n.mask_o.eq(Cat(*ml))
                m.d.comb += self.n.stop_o.eq(Cat(*ms))

        m.d.comb += eq(self.n.data_o, self.process(r_data[mid]))

        return m


class NonCombMultiInPipeline(MultiInControlBase):
    """ A multi-input pipeline block conforming to the Pipeline API

        Attributes:
        -----------
        p.data_i : StageInput, shaped according to ispec
            The pipeline input
        p.data_o : StageOutput, shaped according to ospec
            The pipeline output
        r_data : input_shape according to ispec
            A temporary (buffered) copy of a prior (valid) input.
            This is HELD if the output is not ready.  It is updated
            SYNCHRONOUSLY.
    """

    def __init__(self, stage, p_len, p_mux, maskwid=0, routemask=False):
        MultiInControlBase.__init__(self, p_len=p_len, maskwid=maskwid,
                                          routemask=routemask)
        self.stage = stage
        self.maskwid = maskwid
        self.p_mux = p_mux

        # set up the input and output data
        for i in range(p_len):
            name = 'data_i_%d' % i
            self.p[i].data_i = _spec(stage.ispec, name) # input type
        self.n.data_o = _spec(stage.ospec, 'data_o')

    def process(self, i):
        if hasattr(self.stage, "process"):
            return self.stage.process(i)
        return i

    def elaborate(self, platform):
        m = MultiInControlBase.elaborate(self, platform)

        m.submodules.p_mux = self.p_mux

        # need an array of buffer registers conforming to *input* spec
        r_data = []
        r_busy = []
        p_valid_i = []
        p_len = len(self.p)
        for i in range(p_len):
            name = 'r_%d' % i
            r = _spec(self.stage.ispec, name) # input type
            r_data.append(r)
            r_busy.append(Signal(name="r_busy%d" % i, reset_less=True))
            p_valid_i.append(Signal(name="p_valid_i%d" % i, reset_less=True))
            if hasattr(self.stage, "setup"):
                print ("setup", self, self.stage, r)
                self.stage.setup(m, r)
        if len(r_data) > 1:
            r_data = Array(r_data)
            p_valid_i = Array(p_valid_i)
            r_busy = Array(r_busy)

        nirn = Signal(reset_less=True)
        m.d.comb += nirn.eq(~self.n.ready_i)
        mid = self.p_mux.m_id
        print ("CombMuxIn mid", self, self.stage, self.routemask, mid, p_len)
        for i in range(p_len):
            m.d.comb += r_busy[i].eq(0)
            m.d.comb += n_ready_in[i].eq(1)
            m.d.comb += p_valid_i[i].eq(0)
            m.d.comb += self.p[i].ready_o.eq(n_ready_in[i])
        p = self.p[mid]
        maskedout = Signal(reset_less=True)
        if hasattr(p, "mask_i"):
            m.d.comb += maskedout.eq(p.mask_i & ~p.stop_i)
        else:
            m.d.comb += maskedout.eq(1)
        m.d.comb += p_valid_i[mid].eq(maskedout & self.p_mux.active)
        m.d.comb += self.p[mid].ready_o.eq(~data_valid[mid] | self.n.ready_i)
        m.d.comb += n_ready_in[mid].eq(nirn & data_valid[mid])
        anyvalid = Signal(i, reset_less=True)
        av = []
        for i in range(p_len):
            av.append(data_valid[i])
        anyvalid = Cat(*av)
        m.d.comb += self.n.valid_o.eq(anyvalid.bool())
        m.d.comb += data_valid[mid].eq(p_valid_i[mid] | \
                                    (n_ready_in[mid] ))

        if self.routemask:
            # XXX hack - fixes loop
            m.d.comb += eq(self.n.stop_o, self.p[-1].stop_i)
            for i in range(p_len):
                p = self.p[i]
                vr = Signal(name="vr%d" % i, reset_less=True)
                maskedout = Signal(name="maskedout%d" % i, reset_less=True)
                if hasattr(p, "mask_i"):
                    m.d.comb += maskedout.eq(p.mask_i & ~p.stop_i)
                else:
                    m.d.comb += maskedout.eq(1)
                m.d.comb += vr.eq(maskedout.bool() & p.valid_i & p.ready_o)
                #m.d.comb += vr.eq(p.valid_i & p.ready_o)
                with m.If(vr):
                    m.d.comb += eq(self.n.mask_o, self.p[i].mask_i)
                    m.d.comb += eq(r_data[i], self.p[i].data_i)
        else:
            ml = [] # accumulate output masks
            ms = [] # accumulate output stops
            for i in range(p_len):
                vr = Signal(reset_less=True)
                p = self.p[i]
                vr = Signal(reset_less=True)
                maskedout = Signal(reset_less=True)
                if hasattr(p, "mask_i"):
                    m.d.comb += maskedout.eq(p.mask_i & ~p.stop_i)
                else:
                    m.d.comb += maskedout.eq(1)
                m.d.comb += vr.eq(maskedout.bool() & p.valid_i & p.ready_o)
                with m.If(vr):
                    m.d.comb += eq(r_data[i], self.p[i].data_i)
                if self.maskwid:
                    mlen = len(self.p[i].mask_i)
                    s = mlen*i
                    e = mlen*(i+1)
                    ml.append(Mux(vr, self.p[i].mask_i, Const(0, mlen)))
                    ms.append(self.p[i].stop_i)
            if self.maskwid:
                m.d.comb += self.n.mask_o.eq(Cat(*ml))
                m.d.comb += self.n.stop_o.eq(Cat(*ms))

        m.d.comb += eq(self.n.data_o, self.process(r_data[mid]))

        return m


class CombMuxOutPipe(CombMultiOutPipeline):
    def __init__(self, stage, n_len, maskwid=0, muxidname=None,
                                     routemask=False):
        muxidname = muxidname or "muxid"
        # HACK: stage is also the n-way multiplexer
        CombMultiOutPipeline.__init__(self, stage, n_len=n_len,
                                            n_mux=stage, maskwid=maskwid,
                                            routemask=routemask)

        # HACK: n-mux is also the stage... so set the muxid equal to input muxid
        muxid = getattr(self.p.data_i, muxidname)
        print ("combmuxout", muxidname, muxid)
        stage.m_id = muxid



class InputPriorityArbiter(Elaboratable):
    """ arbitration module for Input-Mux pipe, baed on PriorityEncoder
    """
    def __init__(self, pipe, num_rows):
        self.pipe = pipe
        self.num_rows = num_rows
        self.mmax = int(log(self.num_rows) / log(2))
        self.m_id = Signal(self.mmax, reset_less=True) # multiplex id
        self.active = Signal(reset_less=True)

    def elaborate(self, platform):
        m = Module()

        assert len(self.pipe.p) == self.num_rows, \
                "must declare input to be same size"
        pe = PriorityEncoder(self.num_rows)
        m.submodules.selector = pe

        # connect priority encoder
        in_ready = []
        for i in range(self.num_rows):
            p_valid_i = Signal(reset_less=True)
            if self.pipe.maskwid and not self.pipe.routemask:
                p = self.pipe.p[i]
                maskedout = Signal(reset_less=True)
                m.d.comb += maskedout.eq(p.mask_i & ~p.stop_i)
                m.d.comb += p_valid_i.eq(maskedout.bool() & p.valid_i_test)
            else:
                m.d.comb += p_valid_i.eq(self.pipe.p[i].valid_i_test)
            in_ready.append(p_valid_i)
        m.d.comb += pe.i.eq(Cat(*in_ready)) # array of input "valids"
        m.d.comb += self.active.eq(~pe.n)   # encoder active (one input valid)
        m.d.comb += self.m_id.eq(pe.o)       # output one active input

        return m

    def ports(self):
        return [self.m_id, self.active]



class PriorityCombMuxInPipe(CombMultiInPipeline):
    """ an example of how to use the combinatorial pipeline.
    """

    def __init__(self, stage, p_len=2, maskwid=0, routemask=False):
        p_mux = InputPriorityArbiter(self, p_len)
        CombMultiInPipeline.__init__(self, stage, p_len, p_mux,
                                     maskwid=maskwid, routemask=routemask)


if __name__ == '__main__':

    from nmutil.test.example_buf_pipe import ExampleStage
    dut = PriorityCombMuxInPipe(ExampleStage)
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_combpipe.il", "w") as f:
        f.write(vl)



""" Unit tests for Buffered and Unbuffered pipelines

    contains useful worked examples of how to use the Pipeline API,
    including:

    * Combinatorial Stage "Chaining"
    * class-based data stages
    * nmigen module-based data stages
    * special nmigen module-based data stage, where the stage *is* the module
    * Record-based data stages
    * static-class data stages
    * multi-stage pipelines (and how to connect them)
    * how to *use* the pipelines (see Test5) - how to get data in and out

"""

from nmigen import Module, Signal, Mux, Const, Elaboratable
from nmigen.hdl.rec import Record
from nmigen.compat.sim import run_simulation
from nmigen.cli import verilog, rtlil

from nmutil.test.example_buf_pipe import ExampleBufPipe, ExampleBufPipeAdd
from nmutil.test.example_buf_pipe import ExamplePipeline, UnbufferedPipeline
from nmutil.test.example_buf_pipe import ExampleStageCls
from nmutil.iocontrol import PrevControl, NextControl
from nmutil.stageapi import StageChain, StageCls
from nmutil.singlepipe import ControlBase
from nmutil.singlepipe import UnbufferedPipeline2
from nmutil.singlepipe import SimpleHandshake
from nmutil.singlepipe import BufferedHandshake
from nmutil.singlepipe import PassThroughHandshake
from nmutil.singlepipe import PassThroughStage
from nmutil.singlepipe import FIFOControl
from nmutil.singlepipe import RecordObject
from nmutil.singlepipe import MaskCancellable

from random import randint, seed

import unittest

# seed(4)


def check_o_n_valid(dut, val):
    o_n_valid = yield dut.n.valid_o
    assert o_n_valid == val


def check_o_n_valid2(dut, val):
    o_n_valid = yield dut.n.valid_o
    assert o_n_valid == val


def tbench(dut):
    # yield dut.i_p_rst.eq(1)
    yield dut.n.ready_i.eq(0)
    # yield dut.p.ready_o.eq(0)
    yield
    yield
    # yield dut.i_p_rst.eq(0)
    yield dut.n.ready_i.eq(1)
    yield dut.p.data_i.eq(5)
    yield dut.p.valid_i.eq(1)
    yield

    yield dut.p.data_i.eq(7)
    yield from check_o_n_valid(dut, 0)  # effects of i_p_valid delayed
    yield
    yield from check_o_n_valid(dut, 1)  # ok *now* i_p_valid effect is felt

    yield dut.p.data_i.eq(2)
    yield
    # begin going into "stall" (next stage says ready)
    yield dut.n.ready_i.eq(0)
    yield dut.p.data_i.eq(9)
    yield
    yield dut.p.valid_i.eq(0)
    yield dut.p.data_i.eq(12)
    yield
    yield dut.p.data_i.eq(32)
    yield dut.n.ready_i.eq(1)
    yield
    yield from check_o_n_valid(dut, 1)  # buffer still needs to output
    yield
    yield from check_o_n_valid(dut, 1)  # buffer still needs to output
    yield
    yield from check_o_n_valid(dut, 0)  # buffer outputted, *now* we're done.
    yield


def tbench2(dut):
    # yield dut.p.i_rst.eq(1)
    yield dut.n.ready_i.eq(0)
    # yield dut.p.ready_o.eq(0)
    yield
    yield
    # yield dut.p.i_rst.eq(0)
    yield dut.n.ready_i.eq(1)
    yield dut.p.data_i.eq(5)
    yield dut.p.valid_i.eq(1)
    yield

    yield dut.p.data_i.eq(7)
    # effects of i_p_valid delayed 2 clocks
    yield from check_o_n_valid2(dut, 0)
    yield
    # effects of i_p_valid delayed 2 clocks
    yield from check_o_n_valid2(dut, 0)

    yield dut.p.data_i.eq(2)
    yield
    yield from check_o_n_valid2(dut, 1)  # ok *now* i_p_valid effect is felt
    # begin going into "stall" (next stage says ready)
    yield dut.n.ready_i.eq(0)
    yield dut.p.data_i.eq(9)
    yield
    yield dut.p.valid_i.eq(0)
    yield dut.p.data_i.eq(12)
    yield
    yield dut.p.data_i.eq(32)
    yield dut.n.ready_i.eq(1)
    yield
    yield from check_o_n_valid2(dut, 1)  # buffer still needs to output
    yield
    yield from check_o_n_valid2(dut, 1)  # buffer still needs to output
    yield
    yield from check_o_n_valid2(dut, 1)  # buffer still needs to output
    yield
    yield from check_o_n_valid2(dut, 0)  # buffer outputted, *now* we're done.
    yield
    yield
    yield


class Test3:
    def __init__(self, dut, resultfn):
        self.dut = dut
        self.resultfn = resultfn
        self.data = []
        for i in range(num_tests):
            #data.append(randint(0, 1<<16-1))
            self.data.append(i+1)
        self.i = 0
        self.o = 0

    def send(self):
        while self.o != len(self.data):
            send_range = randint(0, 3)
            for j in range(randint(1, 10)):
                if send_range == 0:
                    send = True
                else:
                    send = randint(0, send_range) != 0
                o_p_ready = yield self.dut.p.ready_o
                if not o_p_ready:
                    yield
                    continue
                if send and self.i != len(self.data):
                    yield self.dut.p.valid_i.eq(1)
                    yield self.dut.p.data_i.eq(self.data[self.i])
                    self.i += 1
                else:
                    yield self.dut.p.valid_i.eq(0)
                yield

    def rcv(self):
        while self.o != len(self.data):
            stall_range = randint(0, 3)
            for j in range(randint(1, 10)):
                stall = randint(0, stall_range) != 0
                yield self.dut.n.ready_i.eq(stall)
                yield
                o_n_valid = yield self.dut.n.valid_o
                i_n_ready = yield self.dut.n.ready_i_test
                if not o_n_valid or not i_n_ready:
                    continue
                data_o = yield self.dut.n.data_o
                self.resultfn(data_o, self.data[self.o], self.i, self.o)
                self.o += 1
                if self.o == len(self.data):
                    break


def resultfn_3(data_o, expected, i, o):
    assert data_o == expected + 1, \
        "%d-%d data %x not match %x\n" \
        % (i, o, data_o, expected)


def data_placeholder():
    data = []
    for i in range(num_tests):
        d = PlaceHolder()
        d.src1 = randint(0, 1 << 16-1)
        d.src2 = randint(0, 1 << 16-1)
        data.append(d)
    return data


def data_dict():
    data = []
    for i in range(num_tests):
        data.append({'src1': randint(0, 1 << 16-1),
                     'src2': randint(0, 1 << 16-1)})
    return data


class Test5:
    def __init__(self, dut, resultfn, data=None, stage_ctl=False):
        self.dut = dut
        self.resultfn = resultfn
        self.stage_ctl = stage_ctl
        if data:
            self.data = data
        else:
            self.data = []
            for i in range(num_tests):
                self.data.append(
                    (randint(0, 1 << 16-1), randint(0, 1 << 16-1)))
        self.i = 0
        self.o = 0

    def send(self):
        while self.o != len(self.data):
            send_range = randint(0, 3)
            for j in range(randint(1, 10)):
                if send_range == 0:
                    send = True
                else:
                    send = randint(0, send_range) != 0
                #send = True
                o_p_ready = yield self.dut.p.ready_o
                if not o_p_ready:
                    yield
                    continue
                if send and self.i != len(self.data):
                    yield self.dut.p.valid_i.eq(1)
                    for v in self.dut.set_input(self.data[self.i]):
                        yield v
                    self.i += 1
                else:
                    yield self.dut.p.valid_i.eq(0)
                yield

    def rcv(self):
        while self.o != len(self.data):
            stall_range = randint(0, 3)
            for j in range(randint(1, 10)):
                ready = randint(0, stall_range) != 0
                #ready = True
                yield self.dut.n.ready_i.eq(ready)
                yield
                o_n_valid = yield self.dut.n.valid_o
                i_n_ready = yield self.dut.n.ready_i_test
                if not o_n_valid or not i_n_ready:
                    continue
                if isinstance(self.dut.n.data_o, Record):
                    data_o = {}
                    dod = self.dut.n.data_o
                    for k, v in dod.fields.items():
                        data_o[k] = yield v
                else:
                    data_o = yield self.dut.n.data_o
                self.resultfn(data_o, self.data[self.o], self.i, self.o)
                self.o += 1
                if self.o == len(self.data):
                    break


class TestMask:
    def __init__(self, dut, resultfn, maskwid, data=None, stage_ctl=False,
                 latching=False):
        self.dut = dut
        self.resultfn = resultfn
        self.stage_ctl = stage_ctl
        self.maskwid = maskwid
        self.latching = latching
        self.latchmode = 0
        if data:
            self.data = data
        else:
            self.data = []
            for i in range(num_tests):
                self.data.append(
                    (randint(0, 1 << 16-1), randint(0, 1 << 16-1)))
        self.i = 0
        self.o = 0

    def send(self):
        while self.o != len(self.data):
            send_range = randint(0, 3)
            for j in range(randint(1, 10)):
                if send_range == 0:
                    send = True
                else:
                    send = randint(0, send_range) != 0
                #send = True
                o_p_ready = yield self.dut.p.ready_o
                if not o_p_ready:
                    yield
                    continue

                if self.latching:
                    latchtest = randint(0, 3) == 0
                    if latchtest:
                        yield self.dut.p.valid_i.eq(0)
                        yield self.dut.p.mask_i.eq(0)
                        # wait for pipeline to flush, then invert state
                        for i in range(10):
                            yield
                        self.latchmode = 1 - self.latchmode
                        yield self.dut.latchmode.eq(self.latchmode)
                        mode = yield self.dut.latchmode
                        print("latching", mode)

                if send and self.i != len(self.data):
                    print("send", self.i, self.data[self.i])
                    yield self.dut.p.valid_i.eq(1)
                    yield self.dut.p.mask_i.eq(1 << self.i)  # XXX TODO
                    for v in self.dut.set_input(self.data[self.i]):
                        yield v
                    self.i += 1
                else:
                    yield self.dut.p.valid_i.eq(0)
                    yield self.dut.p.mask_i.eq(0)  # XXX TODO
                yield

    def rcv(self):
        while self.o != len(self.data):
            stall_range = randint(0, 3)
            for j in range(randint(1, 10)):
                ready = randint(0, stall_range) != 0
                ready = True
                yield self.dut.n.ready_i.eq(ready)
                yield
                o_n_valid = yield self.dut.n.valid_o
                i_n_ready = yield self.dut.n.ready_i_test
                if not o_n_valid or not i_n_ready:
                    continue
                if isinstance(self.dut.n.data_o, Record):
                    data_o = {}
                    dod = self.dut.n.data_o
                    for k, v in dod.fields.items():
                        data_o[k] = yield v
                else:
                    data_o = yield self.dut.n.data_o
                print("recv", self.o, data_o)
                self.resultfn(data_o, self.data[self.o], self.i, self.o)
                self.o += 1
                if self.o == len(self.data):
                    break


def resultfn_5(data_o, expected, i, o):
    res = expected[0] + expected[1]
    assert data_o == res, \
        "%d-%d data %x not match %s\n" \
        % (i, o, data_o, repr(expected))


def tbench4(dut):
    data = []
    for i in range(num_tests):
        #data.append(randint(0, 1<<16-1))
        data.append(i+1)
    i = 0
    o = 0
    while True:
        stall = randint(0, 3) != 0
        send = randint(0, 5) != 0
        yield dut.n.ready_i.eq(stall)
        o_p_ready = yield dut.p.ready_o
        if o_p_ready:
            if send and i != len(data):
                yield dut.p.valid_i.eq(1)
                yield dut.p.data_i.eq(data[i])
                i += 1
            else:
                yield dut.p.valid_i.eq(0)
        yield
        o_n_valid = yield dut.n.valid_o
        i_n_ready = yield dut.n.ready_i_test
        if o_n_valid and i_n_ready:
            data_o = yield dut.n.data_o
            assert data_o == data[o] + 2, "%d-%d data %x not match %x\n" \
                % (i, o, data_o, data[o])
            o += 1
            if o == len(data):
                break

######################################################################
# Test 2 and 4
######################################################################


class ExampleBufPipe2(ControlBase):
    """ Example of how to do chained pipeline stages.
    """

    def elaborate(self, platform):
        m = ControlBase.elaborate(self, platform)

        pipe1 = ExampleBufPipe()
        pipe2 = ExampleBufPipe()

        m.submodules.pipe1 = pipe1
        m.submodules.pipe2 = pipe2

        m.d.comb += self.connect([pipe1, pipe2])

        return m


######################################################################
# Test 9
######################################################################

class ExampleBufPipeChain2(BufferedHandshake):
    """ connects two stages together as a *single* combinatorial stage.
    """

    def __init__(self):
        stage1 = ExampleStageCls()
        stage2 = ExampleStageCls()
        combined = StageChain([stage1, stage2])
        BufferedHandshake.__init__(self, combined)


def data_chain2():
    data = []
    for i in range(num_tests):
        data.append(randint(0, 1 << 16-2))
    return data


def resultfn_9(data_o, expected, i, o):
    res = expected + 2
    assert data_o == res, \
        "%d-%d received data %x not match expected %x\n" \
        % (i, o, data_o, res)


######################################################################
# Test 6 and 10
######################################################################

class SetLessThan(Elaboratable):
    def __init__(self, width, signed):
        self.m = Module()
        self.src1 = Signal((width, signed), name="src1")
        self.src2 = Signal((width, signed), name="src2")
        self.output = Signal(width, name="out")

    def elaborate(self, platform):
        self.m.d.comb += self.output.eq(Mux(self.src1 < self.src2, 1, 0))
        return self.m


class LTStage(StageCls):
    """ module-based stage example
    """

    def __init__(self):
        self.slt = SetLessThan(16, True)

    def ispec(self, name):
        return (Signal(16, name="%s_sig1" % name),
                Signal(16, name="%s_sig2" % name))

    def ospec(self, name):
        return Signal(16, name="%s_out" % name)

    def setup(self, m, i):
        self.o = Signal(16)
        m.submodules.slt = self.slt
        m.d.comb += self.slt.src1.eq(i[0])
        m.d.comb += self.slt.src2.eq(i[1])
        m.d.comb += self.o.eq(self.slt.output)

    def process(self, i):
        return self.o


class LTStageDerived(SetLessThan, StageCls):
    """ special version of a nmigen module where the module is also a stage

        shows that you don't actually need to combinatorially connect
        to the outputs, or add the module as a submodule: just return
        the module output parameter(s) from the Stage.process() function
    """

    def __init__(self):
        SetLessThan.__init__(self, 16, True)

    def ispec(self):
        return (Signal(16), Signal(16))

    def ospec(self):
        return Signal(16)

    def setup(self, m, i):
        m.submodules.slt = self
        m.d.comb += self.src1.eq(i[0])
        m.d.comb += self.src2.eq(i[1])

    def process(self, i):
        return self.output


class ExampleLTPipeline(UnbufferedPipeline):
    """ an example of how to use the unbuffered pipeline.
    """

    def __init__(self):
        stage = LTStage()
        UnbufferedPipeline.__init__(self, stage)


class ExampleLTBufferedPipeDerived(BufferedHandshake):
    """ an example of how to use the buffered pipeline.
    """

    def __init__(self):
        stage = LTStageDerived()
        BufferedHandshake.__init__(self, stage)


def resultfn_6(data_o, expected, i, o):
    res = 1 if expected[0] < expected[1] else 0
    assert data_o == res, \
        "%d-%d data %x not match %s\n" \
        % (i, o, data_o, repr(expected))


######################################################################
# Test 7
######################################################################

class ExampleAddRecordStage(StageCls):
    """ example use of a Record
    """

    record_spec = [('src1', 16), ('src2', 16)]

    def ispec(self):
        """ returns a Record using the specification
        """
        return Record(self.record_spec)

    def ospec(self):
        return Record(self.record_spec)

    def process(self, i):
        """ process the input data, returning a dictionary with key names
            that exactly match the Record's attributes.
        """
        return {'src1': i.src1 + 1,
                'src2': i.src2 + 1}

######################################################################
# Test 11
######################################################################


class ExampleAddRecordPlaceHolderStage(StageCls):
    """ example use of a Record, with a placeholder as the processing result
    """

    record_spec = [('src1', 16), ('src2', 16)]

    def ispec(self):
        """ returns a Record using the specification
        """
        return Record(self.record_spec)

    def ospec(self):
        return Record(self.record_spec)

    def process(self, i):
        """ process the input data, returning a PlaceHolder class instance
            with attributes that exactly match those of the Record.
        """
        o = PlaceHolder()
        o.src1 = i.src1 + 1
        o.src2 = i.src2 + 1
        return o


# a dummy class that may have stuff assigned to instances once created
class PlaceHolder:
    pass


class ExampleAddRecordPipe(UnbufferedPipeline):
    """ an example of how to use the combinatorial pipeline.
    """

    def __init__(self):
        stage = ExampleAddRecordStage()
        UnbufferedPipeline.__init__(self, stage)


def resultfn_7(data_o, expected, i, o):
    res = (expected['src1'] + 1, expected['src2'] + 1)
    assert data_o['src1'] == res[0] and data_o['src2'] == res[1], \
        "%d-%d data %s not match %s\n" \
        % (i, o, repr(data_o), repr(expected))


class ExampleAddRecordPlaceHolderPipe(UnbufferedPipeline):
    """ an example of how to use the combinatorial pipeline.
    """

    def __init__(self):
        stage = ExampleAddRecordPlaceHolderStage()
        UnbufferedPipeline.__init__(self, stage)


def resultfn_11(data_o, expected, i, o):
    res1 = expected.src1 + 1
    res2 = expected.src2 + 1
    assert data_o['src1'] == res1 and data_o['src2'] == res2, \
        "%d-%d data %s not match %s\n" \
        % (i, o, repr(data_o), repr(expected))


######################################################################
# Test 8
######################################################################


class Example2OpClass:
    """ an example of a class used to store 2 operands.
        requires an eq function, to conform with the pipeline stage API
    """

    def __init__(self):
        self.op1 = Signal(16)
        self.op2 = Signal(16)

    def eq(self, i):
        return [self.op1.eq(i.op1), self.op2.eq(i.op2)]


class ExampleAddClassStage(StageCls):
    """ an example of how to use the buffered pipeline, as a class instance
    """

    def ispec(self):
        """ returns an instance of an Example2OpClass.
        """
        return Example2OpClass()

    def ospec(self):
        """ returns an output signal which will happen to contain the sum
            of the two inputs
        """
        return Signal(16, name="add2_out")

    def process(self, i):
        """ process the input data (sums the values in the tuple) and returns it
        """
        return i.op1 + i.op2


class ExampleBufPipeAddClass(BufferedHandshake):
    """ an example of how to use the buffered pipeline, using a class instance
    """

    def __init__(self):
        addstage = ExampleAddClassStage()
        BufferedHandshake.__init__(self, addstage)


class TestInputAdd:
    """ the eq function, called by set_input, needs an incoming object
        that conforms to the Example2OpClass.eq function requirements
        easiest way to do that is to create a class that has the exact
        same member layout (self.op1, self.op2) as Example2OpClass
    """

    def __init__(self, op1, op2):
        self.op1 = op1
        self.op2 = op2


def resultfn_8(data_o, expected, i, o):
    res = expected.op1 + expected.op2  # these are a TestInputAdd instance
    assert data_o == res, \
        "%d-%d data %s res %x not match %s\n" \
        % (i, o, repr(data_o), res, repr(expected))


def data_2op():
    data = []
    for i in range(num_tests):
        data.append(TestInputAdd(randint(0, 1 << 16-1), randint(0, 1 << 16-1)))
    return data


######################################################################
# Test 12
######################################################################

class ExampleStageDelayCls(StageCls, Elaboratable):
    """ an example of how to use the buffered pipeline, in a static class
        fashion
    """

    def __init__(self, valid_trigger=2):
        self.count = Signal(2)
        self.valid_trigger = valid_trigger

    def ispec(self):
        return Signal(16, name="example_input_signal")

    def ospec(self):
        return Signal(16, name="example_output_signal")

    @property
    def d_ready(self):
        """ data is ready to be accepted when this is true
        """
        return (self.count == 1)  # | (self.count == 3)
        return Const(1)

    def d_valid(self, ready_i):
        """ data is valid at output when this is true
        """
        return self.count == self.valid_trigger
        return Const(1)

    def process(self, i):
        """ process the input data and returns it (adds 1)
        """
        return i + 1

    def elaborate(self, platform):
        m = Module()
        m.d.sync += self.count.eq(self.count + 1)
        return m


class ExampleBufDelayedPipe(BufferedHandshake):

    def __init__(self):
        stage = ExampleStageDelayCls(valid_trigger=2)
        BufferedHandshake.__init__(self, stage, stage_ctl=True)

    def elaborate(self, platform):
        m = BufferedHandshake.elaborate(self, platform)
        m.submodules.stage = self.stage
        return m


def data_chain1():
    data = []
    for i in range(num_tests):
        data.append(1 << ((i*3) % 15))
        #data.append(randint(0, 1<<16-2))
        #print (hex(data[-1]))
    return data


def resultfn_12(data_o, expected, i, o):
    res = expected + 1
    assert data_o == res, \
        "%d-%d data %x not match %x\n" \
        % (i, o, data_o, res)


######################################################################
# Test 13
######################################################################

class ExampleUnBufDelayedPipe(BufferedHandshake):

    def __init__(self):
        stage = ExampleStageDelayCls(valid_trigger=3)
        BufferedHandshake.__init__(self, stage, stage_ctl=True)

    def elaborate(self, platform):
        m = BufferedHandshake.elaborate(self, platform)
        m.submodules.stage = self.stage
        return m

######################################################################
# Test 15
######################################################################


class ExampleBufModeAdd1Pipe(SimpleHandshake):

    def __init__(self):
        stage = ExampleStageCls()
        SimpleHandshake.__init__(self, stage)


######################################################################
# Test 16
######################################################################

class ExampleBufModeUnBufPipe(ControlBase):

    def elaborate(self, platform):
        m = ControlBase.elaborate(self, platform)

        pipe1 = ExampleBufModeAdd1Pipe()
        pipe2 = ExampleBufAdd1Pipe()

        m.submodules.pipe1 = pipe1
        m.submodules.pipe2 = pipe2

        m.d.comb += self.connect([pipe1, pipe2])

        return m

######################################################################
# Test 17
######################################################################


class ExampleUnBufAdd1Pipe2(UnbufferedPipeline2):

    def __init__(self):
        stage = ExampleStageCls()
        UnbufferedPipeline2.__init__(self, stage)


######################################################################
# Test 18
######################################################################

class PassThroughTest(PassThroughHandshake):

    def iospecfn(self):
        return Signal(16, name="out")

    def __init__(self):
        stage = PassThroughStage(self.iospecfn)
        PassThroughHandshake.__init__(self, stage)


def resultfn_identical(data_o, expected, i, o):
    res = expected
    assert data_o == res, \
        "%d-%d data %x not match %x\n" \
        % (i, o, data_o, res)


######################################################################
# Test 19
######################################################################

class ExamplePassAdd1Pipe(PassThroughHandshake):

    def __init__(self):
        stage = ExampleStageCls()
        PassThroughHandshake.__init__(self, stage)


class ExampleBufPassThruPipe(ControlBase):

    def elaborate(self, platform):
        m = ControlBase.elaborate(self, platform)

        # XXX currently fails: any other permutation works fine.
        # p1=u,p2=b ok p1=u,p2=u ok p1=b,p2=b ok
        # also fails using UnbufferedPipeline as well
        pipe1 = ExampleBufModeAdd1Pipe()
        pipe2 = ExamplePassAdd1Pipe()

        m.submodules.pipe1 = pipe1
        m.submodules.pipe2 = pipe2

        m.d.comb += self.connect([pipe1, pipe2])

        return m


######################################################################
# Test 20
######################################################################

def iospecfn():
    return Signal(16, name="d_in")


class FIFOTest16(FIFOControl):

    def __init__(self):
        stage = PassThroughStage(iospecfn)
        FIFOControl.__init__(self, 2, stage)


######################################################################
# Test 21
######################################################################

class ExampleFIFOPassThruPipe1(ControlBase):

    def elaborate(self, platform):
        m = ControlBase.elaborate(self, platform)

        pipe1 = FIFOTest16()
        pipe2 = FIFOTest16()
        pipe3 = ExamplePassAdd1Pipe()

        m.submodules.pipe1 = pipe1
        m.submodules.pipe2 = pipe2
        m.submodules.pipe3 = pipe3

        m.d.comb += self.connect([pipe1, pipe2, pipe3])

        return m


######################################################################
# Test 22
######################################################################

class Example2OpRecord(RecordObject):
    def __init__(self):
        RecordObject.__init__(self)
        self.op1 = Signal(16)
        self.op2 = Signal(16)


class ExampleAddRecordObjectStage(StageCls):

    def ispec(self):
        """ returns an instance of an Example2OpRecord.
        """
        return Example2OpRecord()

    def ospec(self):
        """ returns an output signal which will happen to contain the sum
            of the two inputs
        """
        return Signal(16)

    def process(self, i):
        """ process the input data (sums the values in the tuple) and returns it
        """
        return i.op1 + i.op2


class ExampleRecordHandshakeAddClass(SimpleHandshake):

    def __init__(self):
        addstage = ExampleAddRecordObjectStage()
        SimpleHandshake.__init__(self, stage=addstage)


######################################################################
# Test 23
######################################################################

def iospecfnrecord():
    return Example2OpRecord()


class FIFOTestRecordControl(FIFOControl):

    def __init__(self):
        stage = PassThroughStage(iospecfnrecord)
        FIFOControl.__init__(self, 2, stage)


class ExampleFIFORecordObjectPipe(ControlBase):

    def elaborate(self, platform):
        m = ControlBase.elaborate(self, platform)

        pipe1 = FIFOTestRecordControl()
        pipe2 = ExampleRecordHandshakeAddClass()

        m.submodules.pipe1 = pipe1
        m.submodules.pipe2 = pipe2

        m.d.comb += self.connect([pipe1, pipe2])

        return m


######################################################################
# Test 24
######################################################################

class FIFOTestRecordAddStageControl(FIFOControl):

    def __init__(self):
        stage = ExampleAddRecordObjectStage()
        FIFOControl.__init__(self, 2, stage)


######################################################################
# Test 25
######################################################################

class FIFOTestAdd16(FIFOControl):

    def __init__(self):
        stage = ExampleStageCls()
        FIFOControl.__init__(self, 2, stage)


class ExampleFIFOAdd2Pipe(ControlBase):

    def elaborate(self, platform):
        m = ControlBase.elaborate(self, platform)

        pipe1 = FIFOTestAdd16()
        pipe2 = FIFOTestAdd16()

        m.submodules.pipe1 = pipe1
        m.submodules.pipe2 = pipe2

        m.d.comb += self.connect([pipe1, pipe2])

        return m


######################################################################
# Test 26
######################################################################

def iospecfn24():
    return (Signal(16, name="src1"), Signal(16, name="src2"))


class FIFOTest2x16(FIFOControl):

    def __init__(self):
        stage = PassThroughStage(iospecfn2)
        FIFOControl.__init__(self, 2, stage)


######################################################################
# Test 997
######################################################################

class ExampleBufPassThruPipe2(ControlBase):

    def elaborate(self, platform):
        m = ControlBase.elaborate(self, platform)

        # XXX currently fails: any other permutation works fine.
        # p1=u,p2=b ok p1=u,p2=u ok p1=b,p2=b ok
        # also fails using UnbufferedPipeline as well
        #pipe1 = ExampleUnBufAdd1Pipe()
        #pipe2 = ExampleBufAdd1Pipe()
        pipe1 = ExampleBufAdd1Pipe()
        pipe2 = ExamplePassAdd1Pipe()

        m.submodules.pipe1 = pipe1
        m.submodules.pipe2 = pipe2

        m.d.comb += self.connect([pipe1, pipe2])

        return m


######################################################################
# Test 998
######################################################################

class ExampleBufPipe3(ControlBase):
    """ Example of how to do delayed pipeline, where the stage signals
        whether it is ready.
    """

    def elaborate(self, platform):
        m = ControlBase.elaborate(self, platform)

        pipe1 = ExampleBufDelayedPipe()
        pipe2 = ExampleBufPipe()

        m.submodules.pipe1 = pipe1
        m.submodules.pipe2 = pipe2

        m.d.comb += self.connect([pipe1, pipe2])

        return m

######################################################################
# Test 999 - XXX FAILS
# http://bugs.libre-riscv.org/show_bug.cgi?id=57
######################################################################


class ExampleBufAdd1Pipe(BufferedHandshake):

    def __init__(self):
        stage = ExampleStageCls()
        BufferedHandshake.__init__(self, stage)


class ExampleUnBufAdd1Pipe(UnbufferedPipeline):

    def __init__(self):
        stage = ExampleStageCls()
        UnbufferedPipeline.__init__(self, stage)


class ExampleBufUnBufPipe(ControlBase):

    def elaborate(self, platform):
        m = ControlBase.elaborate(self, platform)

        # XXX currently fails: any other permutation works fine.
        # p1=u,p2=b ok p1=u,p2=u ok p1=b,p2=b ok
        # also fails using UnbufferedPipeline as well
        #pipe1 = ExampleUnBufAdd1Pipe()
        #pipe2 = ExampleBufAdd1Pipe()
        pipe1 = ExampleBufAdd1Pipe()
        pipe2 = ExampleUnBufAdd1Pipe()

        m.submodules.pipe1 = pipe1
        m.submodules.pipe2 = pipe2

        m.d.comb += self.connect([pipe1, pipe2])

        return m


######################################################################
# Test 0
######################################################################

class ExampleMaskRecord(RecordObject):
    """ an example of a class used to store 2 operands.
        requires an eq function, to conform with the pipeline stage API
    """

    def __init__(self):
        RecordObject.__init__(self)
        self.src1 = Signal(16)
        self.src2 = Signal(16)

    def __eq(self, i):
        return [self.src1.eq(i.src1), self.src2.eq(i.src2)]


class TestInputMask:
    """ the eq function, called by set_input, needs an incoming object
        that conforms to the Example2OpClass.eq function requirements
        easiest way to do that is to create a class that has the exact
        same member layout (self.op1, self.op2) as Example2OpClass
    """

    def __init__(self, src1, src2):
        self.src1 = src1
        self.src2 = src2

    def __repr__(self):
        return "<TestInputMask %x %x" % (self.src1, self.src2)


class ExampleMaskCancellable(StageCls):

    def ispec(self):
        """ returns an instance of an ExampleMaskRecord.
        """
        return ExampleMaskRecord()

    def ospec(self):
        """ returns the same
        """
        return ExampleMaskRecord()

    def process(self, i):
        """ process the input data: increase op1 and op2
        """
        return TestInputMask(i.src1 + 1, i.src2 + 1)


class MaskCancellablePipe(MaskCancellable):

    """ connects two stages together as a *single* combinatorial stage.
    """

    def __init__(self, dynamic=False, maskwid=16):
        stage1 = ExampleMaskCancellable()
        stage2 = ExampleMaskCancellable()
        combined = StageChain([stage1, stage2])
        MaskCancellable.__init__(self, combined, maskwid, dynamic=dynamic)


class MaskCancellablePipe1(MaskCancellable):

    """ connects a stage to a cancellable pipe with "dynamic" mode on.
    """

    def __init__(self, dynamic=True, maskwid=16):
        stage = ExampleMaskCancellable()
        MaskCancellable.__init__(self, stage, maskwid, dynamic=dynamic)


class MaskCancellableDynamic(ControlBase):

    def __init__(self, maskwid):
        self.maskwid = maskwid
        ControlBase.__init__(self, None, maskwid=maskwid)

    def elaborate(self, platform):
        m = ControlBase.elaborate(self, platform)

        pipe1 = MaskCancellablePipe1(maskwid=self.maskwid)
        pipe2 = MaskCancellablePipe1(maskwid=self.maskwid)

        m.submodules.pipe1 = pipe1
        m.submodules.pipe2 = pipe2

        m.d.comb += self.connect([pipe1, pipe2])

        self.latchmode = Signal()
        m.d.comb += pipe1.latchmode.eq(self.latchmode)
        m.d.comb += pipe2.latchmode.eq(self.latchmode)
        #m.d.comb += self.latchmode.eq(1)

        return m


def data_chain0(n_tests):
    data = []
    for i in range(n_tests):
        data.append(TestInputMask(randint(0, 1 << 16-1),
                                  randint(0, 1 << 16-1)))
    return data


def resultfn_0(data_o, expected, i, o):
    assert data_o['src1'] == expected.src1 + 2, \
        "src1 %x-%x received data no match\n" \
        % (data_o['src1'], expected.src1 + 2)
    assert data_o['src2'] == expected.src2 + 2, \
        "src2 %x-%x received data no match\n" \
        % (data_o['src2'], expected.src2 + 2)


######################################################################
# Unit Tests
######################################################################

num_tests = 10


def test0():
    maskwid = num_tests
    print("test 0")
    dut = MaskCancellablePipe(maskwid)
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        dut.p.data_i.ports() + dut.n.data_o.ports()
    vl = rtlil.convert(dut, ports=ports)
    with open("test_maskchain0.il", "w") as f:
        f.write(vl)
    data = data_chain0(maskwid)
    test = TestMask(dut, resultfn_0, maskwid, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_maskchain0.vcd")


def test0_1():
    maskwid = 32
    print("test 0.1")
    dut = MaskCancellableDynamic(maskwid=maskwid)
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o]  # + \
    #dut.p.data_i.ports() + dut.n.data_o.ports()
    vl = rtlil.convert(dut, ports=ports)
    with open("test_maskchain0_dynamic.il", "w") as f:
        f.write(vl)
    data = data_chain0(maskwid)
    test = TestMask(dut, resultfn_0, maskwid, data=data, latching=True)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_maskchain0_dynamic.vcd")


def notworking1():
    print("test 1")
    dut = ExampleBufPipe()
    run_simulation(dut, tbench(dut), vcd_name="test_bufpipe.vcd")


def notworking2():
    print("test 2")
    dut = ExampleBufPipe2()
    run_simulation(dut, tbench2(dut), vcd_name="test_bufpipe2.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_bufpipe2.il", "w") as f:
        f.write(vl)


def test3():
    print("test 3")
    dut = ExampleBufPipe()
    test = Test3(dut, resultfn_3)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufpipe3.vcd")


def test3_5():
    print("test 3.5")
    dut = ExamplePipeline()
    test = Test3(dut, resultfn_3)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_combpipe3.vcd")


def test4():
    print("test 4")
    dut = ExampleBufPipe2()
    run_simulation(dut, tbench4(dut), vcd_name="test_bufpipe4.vcd")


def test5():
    print("test 5")
    dut = ExampleBufPipeAdd()
    test = Test5(dut, resultfn_5, stage_ctl=True)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufpipe5.vcd")


def test6():
    print("test 6")
    dut = ExampleLTPipeline()
    test = Test5(dut, resultfn_6)
    run_simulation(dut, [test.send(), test.rcv()], vcd_name="test_ltcomb6.vcd")

    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        list(dut.p.data_i) + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_ltcomb_pipe.il", "w") as f:
        f.write(vl)


def test7():
    print("test 7")
    dut = ExampleAddRecordPipe()
    data = data_dict()
    test = Test5(dut, resultfn_7, data=data)
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o,
             dut.p.data_i.src1, dut.p.data_i.src2,
             dut.n.data_o.src1, dut.n.data_o.src2]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_recordcomb_pipe.il", "w") as f:
        f.write(vl)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_addrecord.vcd")


def test8():
    print("test 8")
    dut = ExampleBufPipeAddClass()
    data = data_2op()
    test = Test5(dut, resultfn_8, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufpipe8.vcd")


def test9():
    print("test 9")
    dut = ExampleBufPipeChain2()
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_bufpipechain2.il", "w") as f:
        f.write(vl)

    data = data_chain2()
    test = Test5(dut, resultfn_9, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufpipechain2.vcd")


@unittest.skip("broken")  # FIXME
def test10():
    print("test 10")
    dut = ExampleLTBufferedPipeDerived()
    test = Test5(dut, resultfn_6)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_ltbufpipe10.vcd")
    ports = dut.ports()
    vl = rtlil.convert(dut, ports=ports)
    with open("test_ltbufpipe10.il", "w") as f:
        f.write(vl)


def test11():
    print("test 11")
    dut = ExampleAddRecordPlaceHolderPipe()
    data = data_placeholder()
    test = Test5(dut, resultfn_11, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_addrecord.vcd")


def test12():
    print("test 12")
    dut = ExampleBufDelayedPipe()
    data = data_chain1()
    test = Test5(dut, resultfn_12, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufpipe12.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_bufpipe12.il", "w") as f:
        f.write(vl)


def test13():
    print("test 13")
    dut = ExampleUnBufDelayedPipe()
    data = data_chain1()
    test = Test5(dut, resultfn_12, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_unbufpipe13.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_unbufpipe13.il", "w") as f:
        f.write(vl)


def test15():
    print("test 15")
    dut = ExampleBufModeAdd1Pipe()
    data = data_chain1()
    test = Test5(dut, resultfn_12, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufunbuf15.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_bufunbuf15.il", "w") as f:
        f.write(vl)


def test16():
    print("test 16")
    dut = ExampleBufModeUnBufPipe()
    data = data_chain1()
    test = Test5(dut, resultfn_9, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufunbuf16.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_bufunbuf16.il", "w") as f:
        f.write(vl)


def test17():
    print("test 17")
    dut = ExampleUnBufAdd1Pipe2()
    data = data_chain1()
    test = Test5(dut, resultfn_12, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_unbufpipe17.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_unbufpipe17.il", "w") as f:
        f.write(vl)


def test18():
    print("test 18")
    dut = PassThroughTest()
    data = data_chain1()
    test = Test5(dut, resultfn_identical, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_passthru18.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_passthru18.il", "w") as f:
        f.write(vl)


def test19():
    print("test 19")
    dut = ExampleBufPassThruPipe()
    data = data_chain1()
    test = Test5(dut, resultfn_9, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufpass19.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_bufpass19.il", "w") as f:
        f.write(vl)


def test20():
    print("test 20")
    dut = FIFOTest16()
    data = data_chain1()
    test = Test5(dut, resultfn_identical, data=data)
    run_simulation(dut, [test.send(), test.rcv()], vcd_name="test_fifo20.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_fifo20.il", "w") as f:
        f.write(vl)


def test21():
    print("test 21")
    dut = ExampleFIFOPassThruPipe1()
    data = data_chain1()
    test = Test5(dut, resultfn_12, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_fifopass21.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_fifopass21.il", "w") as f:
        f.write(vl)


def test22():
    print("test 22")
    dut = ExampleRecordHandshakeAddClass()
    data = data_2op()
    test = Test5(dut, resultfn_8, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_addrecord22.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i.op1, dut.p.data_i.op2] + \
        [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_addrecord22.il", "w") as f:
        f.write(vl)


def test23():
    print("test 23")
    dut = ExampleFIFORecordObjectPipe()
    data = data_2op()
    test = Test5(dut, resultfn_8, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_addrecord23.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i.op1, dut.p.data_i.op2] + \
        [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_addrecord23.il", "w") as f:
        f.write(vl)


def test24():
    print("test 24")
    dut = FIFOTestRecordAddStageControl()
    data = data_2op()
    test = Test5(dut, resultfn_8, data=data)
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i.op1, dut.p.data_i.op2] + \
        [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_addrecord24.il", "w") as f:
        f.write(vl)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_addrecord24.vcd")


def test25():
    print("test 25")
    dut = ExampleFIFOAdd2Pipe()
    data = data_chain1()
    test = Test5(dut, resultfn_9, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_add2pipe25.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_add2pipe25.il", "w") as f:
        f.write(vl)


def test997():
    print("test 997")
    dut = ExampleBufPassThruPipe2()
    data = data_chain1()
    test = Test5(dut, resultfn_9, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufpass997.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_bufpass997.il", "w") as f:
        f.write(vl)


@unittest.skip("buggy -- fails due to exceeding step count limit")  # FIXME
def test998():
    print("test 998 (fails, bug)")
    dut = ExampleBufPipe3()
    data = data_chain1()
    test = Test5(dut, resultfn_9, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufpipe14.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_bufpipe14.il", "w") as f:
        f.write(vl)


def test999():
    print("test 999 (expected to fail, which is a bug)")
    dut = ExampleBufUnBufPipe()
    data = data_chain1()
    test = Test5(dut, resultfn_9, data=data)
    run_simulation(dut, [test.send(), test.rcv()],
                   vcd_name="test_bufunbuf999.vcd")
    ports = [dut.p.valid_i, dut.n.ready_i,
             dut.n.valid_o, dut.p.ready_o] + \
        [dut.p.data_i] + [dut.n.data_o]
    vl = rtlil.convert(dut, ports=ports)
    with open("test_bufunbuf999.il", "w") as f:
        f.write(vl)


if __name__ == '__main__':
    test0()
    test0_1()



import os
import re
import shutil
import subprocess
import textwrap
import traceback
import unittest
import warnings
from contextlib import contextmanager

from nmigen.hdl.ir import Fragment
from nmigen.back import rtlil
from nmigen._toolchain import require_tool


__all__ = ["FHDLTestCase"]


class FHDLTestCase(unittest.TestCase):
    def assertRepr(self, obj, repr_str):
        if isinstance(obj, list):
            obj = Statement.cast(obj)
        def prepare_repr(repr_str):
            repr_str = re.sub(r"\s+",   " ",  repr_str)
            repr_str = re.sub(r"\( (?=\()", "(", repr_str)
            repr_str = re.sub(r"\) (?=\))", ")", repr_str)
            return repr_str.strip()
        self.assertEqual(prepare_repr(repr(obj)), prepare_repr(repr_str))

    @contextmanager
    def assertRaises(self, exception, msg=None):
        with super().assertRaises(exception) as cm:
            yield
        if msg is not None:
            # WTF? unittest.assertRaises is completely broken.
            self.assertEqual(str(cm.exception), msg)

    @contextmanager
    def assertRaisesRegex(self, exception, regex=None):
        with super().assertRaises(exception) as cm:
            yield
        if regex is not None:
            # unittest.assertRaisesRegex also seems broken...
            self.assertRegex(str(cm.exception), regex)

    @contextmanager
    def assertWarns(self, category, msg=None):
        with warnings.catch_warnings(record=True) as warns:
            yield
        self.assertEqual(len(warns), 1)
        self.assertEqual(warns[0].category, category)
        if msg is not None:
            self.assertEqual(str(warns[0].message), msg)

    def assertFormal(self, spec, mode="bmc", depth=1, solver=""):
        caller, *_ = traceback.extract_stack(limit=2)
        spec_root, _ = os.path.splitext(caller.filename)
        spec_dir = os.path.dirname(spec_root)
        spec_name = "{}_{}".format(
            os.path.basename(spec_root).replace("test_", "spec_"),
            caller.name.replace("test_", "")
        )

        # The sby -f switch seems not fully functional when sby is
        # reading from stdin.
        if os.path.exists(os.path.join(spec_dir, spec_name)):
            shutil.rmtree(os.path.join(spec_dir, spec_name))

        if mode == "hybrid":
            # A mix of BMC and k-induction, as per personal
            # communication with Clifford Wolf.
            script = "setattr -unset init w:* a:nmigen.sample_reg %d"
            mode   = "bmc"
        else:
            script = ""

        config = textwrap.dedent("""\
        [options]
        mode {mode}
        depth {depth}
        wait on

        [engines]
        smtbmc {solver}

        [script]
        read_ilang top.il
        prep
        {script}

        [file top.il]
        {rtlil}
        """).format(
            mode=mode,
            depth=depth,
            solver=solver,
            script=script,
            rtlil=rtlil.convert(Fragment.get(spec, platform="formal"))
        )
        with subprocess.Popen([require_tool("sby"), "-f", "-d", spec_name],
                              cwd=spec_dir,
                              universal_newlines=True,
                              stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE) as proc:
            stdout, stderr = proc.communicate(config)
            if proc.returncode != 0:
                self.fail("Formal verification failed:\n" + stdout)




from random import randint
from math import log
from nmigen import Module, Signal, Cat
from nmigen.compat.sim import run_simulation
from nmigen.cli import verilog, rtlil

from nmutil.singlepipe import PassThroughStage
from nmutil.multipipe import (CombMultiInPipeline, PriorityCombMuxInPipe)

from . import StepLimiter
import unittest


class PassData:
    def __init__(self):
        self.muxid = Signal(2, reset_less=True)
        self.idx = Signal(6, reset_less=True)
        self.data = Signal(16, reset_less=True)

    def eq(self, i):
        return [self.muxid.eq(i.muxid), self.idx.eq(i.idx), self.data.eq(i.data)]

    def ports(self):
        return [self.muxid, self.idx, self.data]


def tbench(dut):
    stb = yield dut.out_op.stb
    assert stb == 0
    ack = yield dut.out_op.ack
    assert ack == 0

    # set row 1 input 0
    yield dut.rs[1].in_op[0].eq(5)
    yield dut.rs[1].stb.eq(0b01)  # strobe indicate 1st op ready
    # yield dut.rs[1].ack.eq(1)
    yield

    # check row 1 output (should be inactive)
    decode = yield dut.rs[1].out_decode
    assert decode == 0
    if False:
        op0 = yield dut.rs[1].out_op[0]
        op1 = yield dut.rs[1].out_op[1]
        assert op0 == 0 and op1 == 0

    # output should be inactive
    out_stb = yield dut.out_op.stb
    assert out_stb == 1

    # set row 0 input 1
    yield dut.rs[1].in_op[1].eq(6)
    yield dut.rs[1].stb.eq(0b11)  # strobe indicate both ops ready

    # set acknowledgement of output... takes 1 cycle to respond
    yield dut.out_op.ack.eq(1)
    yield
    yield dut.out_op.ack.eq(0)  # clear ack on output
    yield dut.rs[1].stb.eq(0)  # clear row 1 strobe

    # output strobe should be active, MID should be 0 until "ack" is set...
    out_stb = yield dut.out_op.stb
    assert out_stb == 1
    out_muxid = yield dut.muxid
    assert out_muxid == 0

    # ... and output should not yet be passed through either
    op0 = yield dut.out_op.v[0]
    op1 = yield dut.out_op.v[1]
    assert op0 == 0 and op1 == 0

    # wait for out_op.ack to activate...
    yield dut.rs[1].stb.eq(0b00)  # set row 1 strobes to zero
    yield

    # *now* output should be passed through
    op0 = yield dut.out_op.v[0]
    op1 = yield dut.out_op.v[1]
    assert op0 == 5 and op1 == 6

    # set row 2 input
    yield dut.rs[2].in_op[0].eq(3)
    yield dut.rs[2].in_op[1].eq(4)
    yield dut.rs[2].stb.eq(0b11)  # strobe indicate 1st op ready
    yield dut.out_op.ack.eq(1)  # set output ack
    yield
    yield dut.rs[2].stb.eq(0)  # clear row 2 strobe
    yield dut.out_op.ack.eq(0)  # set output ack
    yield
    op0 = yield dut.out_op.v[0]
    op1 = yield dut.out_op.v[1]
    assert op0 == 3 and op1 == 4, "op0 %d op1 %d" % (op0, op1)
    out_muxid = yield dut.muxid
    assert out_muxid == 2

    # set row 0 and 3 input
    yield dut.rs[0].in_op[0].eq(9)
    yield dut.rs[0].in_op[1].eq(8)
    yield dut.rs[0].stb.eq(0b11)  # strobe indicate 1st op ready
    yield dut.rs[3].in_op[0].eq(1)
    yield dut.rs[3].in_op[1].eq(2)
    yield dut.rs[3].stb.eq(0b11)  # strobe indicate 1st op ready

    # set acknowledgement of output... takes 1 cycle to respond
    yield dut.out_op.ack.eq(1)
    yield
    yield dut.rs[0].stb.eq(0)  # clear row 1 strobe
    yield
    out_muxid = yield dut.muxid
    assert out_muxid == 0, "out muxid %d" % out_muxid

    yield
    yield dut.rs[3].stb.eq(0)  # clear row 1 strobe
    yield dut.out_op.ack.eq(0)  # clear ack on output
    yield
    out_muxid = yield dut.muxid
    assert out_muxid == 3, "out muxid %d" % out_muxid


class InputTest:
    def __init__(self, dut):
        self.dut = dut
        self.di = {}
        self.do = {}
        self.tlen = 10
        for muxid in range(dut.num_rows):
            self.di[muxid] = {}
            self.do[muxid] = {}
            for i in range(self.tlen):
                self.di[muxid][i] = randint(0, 100) + (muxid << 8)
                self.do[muxid][i] = self.di[muxid][i]

    def send(self, muxid):
        for i in range(self.tlen):
            op2 = self.di[muxid][i]
            rs = self.dut.p[muxid]
            yield rs.valid_i.eq(1)
            yield rs.data_i.data.eq(op2)
            yield rs.data_i.idx.eq(i)
            yield rs.data_i.muxid.eq(muxid)
            yield
            o_p_ready = yield rs.ready_o
            step_limiter = StepLimiter(10000)
            while not o_p_ready:
                step_limiter.step()
                yield
                o_p_ready = yield rs.ready_o

            print("send", muxid, i, hex(op2))

            yield rs.valid_i.eq(0)
            # wait random period of time before queueing another value
            for i in range(randint(0, 3)):
                yield

        yield rs.valid_i.eq(0)
        # wait random period of time before queueing another value
        # for i in range(randint(0, 3)):
        #    yield

        #send_range = randint(0, 3)
        # if send_range == 0:
        #    send = True
        # else:
        #    send = randint(0, send_range) != 0

    def rcv(self):
        for _ in StepLimiter(10000):
            #stall_range = randint(0, 3)
            # for j in range(randint(1,10)):
            #    stall = randint(0, stall_range) != 0
            #    yield self.dut.n[0].ready_i.eq(stall)
            #    yield
            n = self.dut.n
            yield n.ready_i.eq(1)
            yield
            o_n_valid = yield n.valid_o
            i_n_ready = yield n.ready_i
            if not o_n_valid or not i_n_ready:
                continue

            muxid = yield n.data_o.muxid
            out_i = yield n.data_o.idx
            out_v = yield n.data_o.data

            print("recv", muxid, out_i, hex(out_v))

            # see if this output has occurred already, delete it if it has
            assert out_i in self.do[muxid], "out_i %d not in array %s" % \
                (out_i, repr(self.do[muxid]))
            assert self.do[muxid][out_i] == out_v  # pass-through data
            del self.do[muxid][out_i]

            # check if there's any more outputs
            zerolen = True
            for (k, v) in self.do.items():
                if v:
                    zerolen = False
            if zerolen:
                break


class TestPriorityMuxPipe(PriorityCombMuxInPipe):
    def __init__(self):
        self.num_rows = 4
        def iospecfn(): return PassData()
        stage = PassThroughStage(iospecfn)
        PriorityCombMuxInPipe.__init__(self, stage, p_len=self.num_rows)


@unittest.skip("disabled for now: logic loop")  # FIXME
def test1():
    dut = TestPriorityMuxPipe()
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_inputgroup_multi.il", "w") as f:
        f.write(vl)
    #run_simulation(dut, tbench(dut), vcd_name="test_inputgroup.vcd")

    test = InputTest(dut)
    run_simulation(dut, [test.send(1), test.send(0),
                         test.send(3), test.send(2),
                         test.rcv()],
                   vcd_name="test_inputgroup_multi.vcd")


if __name__ == '__main__':
    test1()



""" IO Control API

    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


    Associated development bugs:
    * http://bugs.libre-riscv.org/show_bug.cgi?id=538
    * http://bugs.libre-riscv.org/show_bug.cgi?id=148
    * http://bugs.libre-riscv.org/show_bug.cgi?id=64
    * http://bugs.libre-riscv.org/show_bug.cgi?id=57

    Important: see Stage API (stageapi.py) in combination with below

    Main classes: PrevControl and NextControl.

    These classes manage the data and the synchronisation state
    to the previous and next stage, respectively.  ready/valid
    signals are used by the Pipeline classes to tell if data
    may be safely passed from stage to stage.

    The connection from one stage to the next is carried out with
    NextControl.connect_to_next.  It is *not* necessary to have
    a PrevControl.connect_to_prev because it is functionally
    directly equivalent to prev->next->connect_to_next.
"""

from nmigen import Signal, Cat, Const, Module, Value, Elaboratable
from nmigen.cli import verilog, rtlil
from nmigen.hdl.rec import Record
from nmigen import tracer

from collections.abc import Sequence, Iterable
from collections import OrderedDict

from nmutil import nmoperator


class Object:
    def __init__(self):
        self.fields = OrderedDict()

    def __setattr__(self, k, v):
        print ("kv", k, v)
        if (k.startswith('_') or k in ["fields", "name", "src_loc"] or
           k in dir(Object) or "fields" not in self.__dict__):
            return object.__setattr__(self, k, v)
        self.fields[k] = v

    def __getattr__(self, k):
        if k in self.__dict__:
            return object.__getattr__(self, k)
        try:
            return self.fields[k]
        except KeyError as e:
            raise AttributeError(e)

    def __iter__(self):
        for x in self.fields.values():  # OrderedDict so order is preserved
            if isinstance(x, Iterable):
                yield from x
            else:
                yield x

    def eq(self, inp):
        res = []
        for (k, o) in self.fields.items():
            i = getattr(inp, k)
            print ("eq", o, i)
            rres = o.eq(i)
            if isinstance(rres, Sequence):
                res += rres
            else:
                res.append(rres)
        print (res)
        return res

    def ports(self): # being called "keys" would be much better
        return list(self)


def add_prefix_to_record_signals(prefix, record):
    """recursively hunt through Records, modifying names to add a prefix
    """
    for key, val in record.fields.items():
        if isinstance(val, Signal):
            val.name = prefix + val.name
        elif isinstance(val, Record):
            add_prefix_to_record_signals(prefix, val)


class RecordObject(Record):
    def __init__(self, layout=None, name=None):
        #if name is None:
        #    name = tracer.get_var_name(depth=2, default="$ro")
        Record.__init__(self, layout=layout or [], name=name)


    def __setattr__(self, k, v):
        #print(f"RecordObject setattr({k}, {v})")
        #print (dir(Record))
        if (k.startswith('_') or k in ["fields", "name", "src_loc"] or
           k in dir(Record) or "fields" not in self.__dict__):
            return object.__setattr__(self, k, v)

        if self.name is None:
            prefix = ""
        else:
            prefix = self.name + "_"
        # Prefix the signal name with the name of the recordobject
        if isinstance(v, Signal):
            #print (self, self.name, v.name)
            v.name = prefix + v.name
        elif isinstance(v, Record):
            add_prefix_to_record_signals(prefix, v)

        self.fields[k] = v
        #print ("RecordObject setattr", k, v)
        if isinstance(v, Record):
            newlayout = {k: (k, v.layout)}
        elif isinstance(v, Value):
            newlayout = {k: (k, v.shape())}
        else:
            newlayout = {k: (k, nmoperator.shape(v))}
        self.layout.fields.update(newlayout)

    def __iter__(self):
        for x in self.fields.values(): # remember: fields is an OrderedDict
            if hasattr(x, 'ports'):
                yield from x.ports()
            elif isinstance(x, Record):
                for f in x.fields.values():
                    yield f
            elif isinstance(x, Iterable):
                yield from x           # a bit like flatten (nmigen.tools)
            else:
                yield x

    def ports(self): # would be better being called "keys"
        return list(self)


class PrevControl(Elaboratable):
    """ contains signals that come *from* the previous stage (both in and out)
        * valid_i: previous stage indicating all incoming data is valid.
                   may be a multi-bit signal, where all bits are required
                   to be asserted to indicate "valid".
        * ready_o: output to next stage indicating readiness to accept data
        * data_i : an input - MUST be added by the USER of this class
    """

    def __init__(self, i_width=1, stage_ctl=False, maskwid=0, offs=0):
        self.stage_ctl = stage_ctl
        self.maskwid = maskwid
        if maskwid:
            self.mask_i = Signal(maskwid)                # prev   >>in  self
            self.stop_i = Signal(maskwid)                # prev   >>in  self
        self.valid_i = Signal(i_width, name="p_valid_i") # prev   >>in  self
        self._ready_o = Signal(name="p_ready_o")         # prev   <<out self
        self.data_i = None # XXX MUST BE ADDED BY USER
        if stage_ctl:
            self.s_ready_o = Signal(name="p_s_o_rdy")    # prev   <<out self
        self.trigger = Signal(reset_less=True)

    @property
    def ready_o(self):
        """ public-facing API: indicates (externally) that stage is ready
        """
        if self.stage_ctl:
            return self.s_ready_o # set dynamically by stage
        return self._ready_o      # return this when not under dynamic control

    def _connect_in(self, prev, direct=False, fn=None,
                    do_data=True, do_stop=True):
        """ internal helper function to connect stage to an input source.
            do not use to connect stage-to-stage!
        """
        valid_i = prev.valid_i if direct else prev.valid_i_test
        res = [self.valid_i.eq(valid_i),
               prev.ready_o.eq(self.ready_o)]
        if self.maskwid:
            res.append(self.mask_i.eq(prev.mask_i))
            if do_stop:
                res.append(self.stop_i.eq(prev.stop_i))
        if do_data is False:
            return res
        data_i = fn(prev.data_i) if fn is not None else prev.data_i
        return res + [nmoperator.eq(self.data_i, data_i)]

    @property
    def valid_i_test(self):
        vlen = len(self.valid_i)
        if vlen > 1:
            # multi-bit case: valid only when valid_i is all 1s
            all1s = Const(-1, (len(self.valid_i), False))
            valid_i = (self.valid_i == all1s)
        else:
            # single-bit valid_i case
            valid_i = self.valid_i

        # when stage indicates not ready, incoming data
        # must "appear" to be not ready too
        if self.stage_ctl:
            valid_i = valid_i & self.s_ready_o

        return valid_i

    def elaborate(self, platform):
        m = Module()
        m.d.comb += self.trigger.eq(self.valid_i_test & self.ready_o)
        return m

    def eq(self, i):
        res = [nmoperator.eq(self.data_i, i.data_i),
                self.ready_o.eq(i.ready_o),
                self.valid_i.eq(i.valid_i)]
        if self.maskwid:
            res.append(self.mask_i.eq(i.mask_i))
        return res

    def __iter__(self):
        yield self.valid_i
        yield self.ready_o
        if self.maskwid:
            yield self.mask_i
            yield self.stop_i
        if hasattr(self.data_i, "ports"):
            yield from self.data_i.ports()
        elif (isinstance(self.data_i, Sequence) or
              isinstance(self.data_i, Iterable)):
            yield from self.data_i
        else:
            yield self.data_i

    def ports(self):
        return list(self)


class NextControl(Elaboratable):
    """ contains the signals that go *to* the next stage (both in and out)
        * valid_o: output indicating to next stage that data is valid
        * ready_i: input from next stage indicating that it can accept data
        * data_o : an output - MUST be added by the USER of this class
    """
    def __init__(self, stage_ctl=False, maskwid=0):
        self.stage_ctl = stage_ctl
        self.maskwid = maskwid
        if maskwid:
            self.mask_o = Signal(maskwid)       # self out>>  next
            self.stop_o = Signal(maskwid)       # self out>>  next
        self.valid_o = Signal(name="n_valid_o") # self out>>  next
        self.ready_i = Signal(name="n_ready_i") # self <<in   next
        self.data_o = None # XXX MUST BE ADDED BY USER
        #if self.stage_ctl:
        self.d_valid = Signal(reset=1) # INTERNAL (data valid)
        self.trigger = Signal(reset_less=True)

    @property
    def ready_i_test(self):
        if self.stage_ctl:
            return self.ready_i & self.d_valid
        return self.ready_i

    def connect_to_next(self, nxt, do_data=True, do_stop=True):
        """ helper function to connect to the next stage data/valid/ready.
            data/valid is passed *TO* nxt, and ready comes *IN* from nxt.
            use this when connecting stage-to-stage

            note: a "connect_from_prev" is completely unnecessary: it's
            just nxt.connect_to_next(self)
        """
        res = [nxt.valid_i.eq(self.valid_o),
               self.ready_i.eq(nxt.ready_o)]
        if self.maskwid:
            res.append(nxt.mask_i.eq(self.mask_o))
            if do_stop:
                res.append(nxt.stop_i.eq(self.stop_o))
        if do_data:
            res.append(nmoperator.eq(nxt.data_i, self.data_o))
        print ("connect to next", self, self.maskwid, nxt.data_i,
                                  do_data, do_stop)
        return res

    def _connect_out(self, nxt, direct=False, fn=None,
                     do_data=True, do_stop=True):
        """ internal helper function to connect stage to an output source.
            do not use to connect stage-to-stage!
        """
        ready_i = nxt.ready_i if direct else nxt.ready_i_test
        res = [nxt.valid_o.eq(self.valid_o),
               self.ready_i.eq(ready_i)]
        if self.maskwid:
            res.append(nxt.mask_o.eq(self.mask_o))
            if do_stop:
                res.append(nxt.stop_o.eq(self.stop_o))
        if not do_data:
            return res
        data_o = fn(nxt.data_o) if fn is not None else nxt.data_o
        return res + [nmoperator.eq(data_o, self.data_o)]

    def elaborate(self, platform):
        m = Module()
        m.d.comb += self.trigger.eq(self.ready_i_test & self.valid_o)
        return m

    def __iter__(self):
        yield self.ready_i
        yield self.valid_o
        if self.maskwid:
            yield self.mask_o
            yield self.stop_o
        if hasattr(self.data_o, "ports"):
            yield from self.data_o.ports()
        elif (isinstance(self.data_o, Sequence) or
              isinstance(self.data_o, Iterable)):
            yield from self.data_o
        else:
            yield self.data_o

    def ports(self):
        return list(self)




from random import randint
from math import log
from nmigen import Module, Signal, Cat, Elaboratable
from nmigen.compat.sim import run_simulation
from nmigen.cli import verilog, rtlil

from nmutil.multipipe import CombMuxOutPipe
from nmutil.singlepipe import SimpleHandshake, PassThroughHandshake, RecordObject


class PassInData(RecordObject):
    def __init__(self):
        RecordObject.__init__(self)
        self.muxid = Signal(2, reset_less=True)
        self.data = Signal(16, reset_less=True)


class PassThroughStage:

    def ispec(self):
        return PassInData()

    def ospec(self, name):
        return Signal(16, name="%s_dout" % name, reset_less=True)
                
    def process(self, i):
        return i.data


class PassThroughDataStage:
    def ispec(self):
        return PassInData()
    def ospec(self):
        return self.ispec() # same as ospec

    def process(self, i):
        return i # pass-through



class PassThroughPipe(PassThroughHandshake):
    def __init__(self):
        PassThroughHandshake.__init__(self, PassThroughDataStage())


class OutputTest:
    def __init__(self, dut):
        self.dut = dut
        self.di = []
        self.do = {}
        self.tlen = 10
        for i in range(self.tlen * dut.num_rows):
            if i < dut.num_rows:
                muxid = i
            else:
                muxid = randint(0, dut.num_rows-1)
            data = randint(0, 255) + (muxid<<8)
            if muxid not in self.do:
                self.do[muxid] = []
            self.di.append((data, muxid))
            self.do[muxid].append(data)

    def send(self):
        for i in range(self.tlen * self.dut.num_rows):
            op2 = self.di[i][0]
            muxid = self.di[i][1]
            rs = self.dut.p
            yield rs.valid_i.eq(1)
            yield rs.data_i.data.eq(op2)
            yield rs.data_i.muxid.eq(muxid)
            yield
            o_p_ready = yield rs.ready_o
            while not o_p_ready:
                yield
                o_p_ready = yield rs.ready_o

            print ("send", muxid, i, hex(op2))

            yield rs.valid_i.eq(0)
            # wait random period of time before queueing another value
            for i in range(randint(0, 3)):
                yield

        yield rs.valid_i.eq(0)

    def rcv(self, muxid):
        out_i = 0
        count = 0
        stall_range = randint(0, 3)
        while out_i != len(self.do[muxid]):
            count += 1
            assert count != 2000, "timeout: too long"
            n = self.dut.n[muxid]
            yield n.ready_i.eq(1)
            yield
            o_n_valid = yield n.valid_o
            i_n_ready = yield n.ready_i
            if not o_n_valid or not i_n_ready:
                continue

            out_v = yield n.data_o

            print ("recv", muxid, out_i, hex(out_v))

            assert self.do[muxid][out_i] == out_v # pass-through data

            out_i += 1

            if randint(0, 5) == 0:
                stall_range = randint(0, 3)
            stall = randint(0, stall_range) != 0
            if stall:
                yield n.ready_i.eq(0)
                for i in range(stall_range):
                    yield


class TestPriorityMuxPipe(CombMuxOutPipe):
    def __init__(self, num_rows):
        self.num_rows = num_rows
        stage = PassThroughStage()
        CombMuxOutPipe.__init__(self, stage, n_len=self.num_rows)


class TestSyncToPriorityPipe(Elaboratable):
    def __init__(self):
        self.num_rows = 4
        self.pipe = PassThroughPipe()
        self.muxpipe = TestPriorityMuxPipe(self.num_rows)

        self.p = self.pipe.p
        self.n = self.muxpipe.n

    def elaborate(self, platform):
        m = Module()
        m.submodules.pipe = self.pipe
        m.submodules.muxpipe = self.muxpipe
        m.d.comb += self.pipe.n.connect_to_next(self.muxpipe.p)
        return m

    def ports(self):
        res = [self.p.valid_i, self.p.ready_o] + \
                self.p.data_i.ports()
        for i in range(len(self.n)):
            res += [self.n[i].ready_i, self.n[i].valid_o] + \
                    [self.n[i].data_o]
                    #self.n[i].data_o.ports()
        return res


def test1():
    dut = TestSyncToPriorityPipe()
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_outmux_pipe.il", "w") as f:
        f.write(vl)

    test = OutputTest(dut)
    run_simulation(dut, [test.rcv(1), test.rcv(0),
                         test.rcv(3), test.rcv(2),
                         test.send()],
                   vcd_name="test_outmux_pipe.vcd")

if __name__ == '__main__':
    test1()



"""
    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+
"""

from nmigen import Signal, Cat

# TODO: turn this into a module?
def byte_reverse(m, name, data, length):
    """byte_reverse: unlike nmigen word_select this takes a dynamic length

    nmigen Signal.word_select may only take a fixed length.  we need
    bigendian byte-reverse, half-word reverse, word and dword reverse.
    """
    comb = m.d.comb
    data_r = Signal.like(data, name=name)

    # if length is a static integer, we do not require a Case statement
    if isinstance(length, int):
        j = length
        rev = []
        for i in range(j):
            dest = data_r.word_select(i, 8)
            rev.append(data.word_select(j-1-i, 8))
        comb += data_r.eq(Cat(*rev))
        return data_r

    # Switch statement needed: dynamic length had better be = 1,2,4 or 8
    with m.Switch(length):
        for j in [1,2,4,8]:
            with m.Case(j):
                rev = []
                for i in range(j):
                    rev.append(data.word_select(j-1-i, 8))
                comb += data_r.eq(Cat(*rev))
    return data_r




"""
    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


"""
from nmigen import Repl, Cat, Const


def exts(exts_data, width, fullwidth):
    exts_data = exts_data[0:width]
    topbit = exts_data[-1]
    signbits = Repl(topbit, fullwidth-width)
    return Cat(exts_data, signbits)


def extz(exts_data, width, fullwidth):
    exts_data = exts_data[0:width]
    topbit = Const(0)
    signbits = Repl(topbit, fullwidth-width)
    return Cat(exts_data, signbits)





"""
    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


"""
from vcd.gtkw import GTKWSave, GTKWColor
from math import log2


def write_gtkw(gtkw_name, vcd_name, gtkw_dom, gtkw_style=None,
               module=None, loc=None, color=None, base=None,
               zoom=None, marker=-1, clk_period=1e-6,
               time_resolution_unit="ps"):
    """ Write a GTKWave document according to the supplied style and DOM.

    :param gtkw_name: name of the generated GTKWave document
    :param vcd_name: name of the waveform file
    :param gtkw_dom: DOM style description for the trace pane
    :param gtkw_style: style for signals, classes and groups
    :param module: default module
    :param color: default trace color
    :param base: default numerical base
    :param loc: source code location to include as a comment
    :param zoom: initial zoom level, in GTKWave format. Can also be "formal"
                 when the file comes from a formal engine
    :param marker: initial location of a marker
    :param clk_period: clock period in seconds, helping
                       to set a reasonable initial zoom level.
                       Use together with ``time_resolution_unit``.
    :param time_resolution_unit: use "ps" or "ns". Derived from the units of
                                 the "timescale" on the VCD file. Used with
                                 "clk_period" to set a default zoom level

    **gtkw_style format**

    Syntax: ``{selector: {attribute: value, ...}, ...}``

    "selector" can be a signal, class or group

    Signal groups propagate most attributes to their children

    Attribute choices:

    * module: absolute path of the current module
    * submodule: same as above, but relative
    * color: trace color
    * base: numerical base for value display
    * display: alternate text to display in the signal pane
    * comment: comment to display in the signal pane
    * bit: select a bit from a wide signal. MSB is zero, unfortunately
    * closed (for groups): this group starts closed

    **gtkw_dom format**

    Syntax: ``[signal, (signal, class), (group, [children]), comment, ...]``

    The DOM is a list of nodes.

    Nodes are signals, signal groups or comments.

    * signals are strings, or tuples: ``(signal name, class, class, ...)``
    * signal groups are tuples: ``(group name, class, class, ..., [nodes])``
    * comments are: ``{'comment': 'comment string'}``

    In place of a class name, an inline class description can be used.
    ``(signal, {attribute: value, ...}, ...)``

    An anonymous group can be used to apply a style to a group of signals.
    ``({attribute: value}, [signal, signal, ...])``
    """
    colors = {
        'blue': GTKWColor.blue,
        'cycle': GTKWColor.cycle,
        'green': GTKWColor.green,
        'indigo': GTKWColor.indigo,
        'normal': GTKWColor.normal,
        'orange': GTKWColor.orange,
        'red': GTKWColor.red,
        'violet': GTKWColor.violet,
        'yellow': GTKWColor.yellow,
    }

    with open(gtkw_name, "wt") as gtkw_file:
        gtkw = GTKWSave(gtkw_file)
        if loc is not None:
            gtkw.comment("Auto-generated by " + loc)
        gtkw.dumpfile(vcd_name)
        # set a reasonable zoom level
        # also, move the marker to an interesting place
        if zoom == "formal":
            # output from formal engines looks good at this zoom level
            zoom = -6.3
        if zoom is None:
            zoom = -42.8 - log2(clk_period)
            # base zoom level is affected by time resolution units
            if time_resolution_unit == "ns":
                zoom = zoom + log2(1e3)
        gtkw.zoom_markers(zoom, marker)

        # create an empty style, if needed
        if gtkw_style is None:
            gtkw_style = dict()

        # create an empty root selector, if needed
        root_style = gtkw_style.get('', dict())

        # apply styles to the root selector, if provided
        if module is not None:
            root_style['module'] = module
        if color is not None:
            root_style['color'] = color
        if base is not None:
            root_style['base'] = base
        # base cannot be None, use 'hex' by default
        if root_style.get('base') is None:
            root_style['base'] = 'hex'

        # recursively walk the DOM
        def walk(dom, style):
            for node in dom:
                node_name = None
                children = None
                # copy the style from the parent
                node_style = style.copy()
                # node is a signal name string
                if isinstance(node, str):
                    node_name = node
                    # apply style from node name, if specified
                    if node_name in gtkw_style:
                        node_style.update(gtkw_style[node_name])
                # node is a tuple
                # could be a signal or a group
                elif isinstance(node, tuple):
                    node_name = node[0]
                    # collect styles from the selectors
                    # order goes from the most specific to most generic
                    # which means earlier selectors override later ones
                    for selector in reversed(node):
                        # update the node style from the selector
                        if isinstance(selector, str):
                            if selector in gtkw_style:
                                node_style.update(gtkw_style[selector])
                        # apply an inline style description
                        elif isinstance(selector, dict):
                            node_style.update(selector)
                    # node is a group if it has a child list
                    if isinstance(node[-1], list):
                        children = node[-1]
                # comment
                elif isinstance(node, dict):
                    if 'comment' in node:
                        gtkw.blank(node['comment'])
                # merge the submodule into the module path
                if 'submodule' in node_style:
                    node_module = node_style['submodule']
                    if 'module' in node_style:
                        node_top_module = node_style['module']
                        node_module = node_top_module + '.' + node_module
                    # update the module path
                    node_style['module'] = node_module
                    # don't propagate this attribute to children
                    del node_style['submodule']
                # emit the group delimiters and walk over the child list
                if children is not None:
                    # see whether this group starts closed
                    closed = False
                    if 'closed' in node_style:
                        closed = node_style['closed']
                        del node_style['closed']  # do not inherit
                    # only emit a group if it has a name
                    if isinstance(node_name, str):
                        gtkw.begin_group(node_name, closed)
                    # pass on the group style to its children
                    walk(children, node_style)
                    if isinstance(node_name, str):
                        gtkw.end_group(node_name)
                # emit a trace, if the node is a signal
                elif node_name is not None:
                    signal_name = node_name
                    # prepend module name to signal
                    if 'module' in node_style:
                        node_module = node_style['module']
                        if node_module is not None:
                            signal_name = node_module + '.' + signal_name
                    node_color = colors.get(node_style.get('color'))
                    node_base = node_style.get('base')
                    display = node_style.get('display')
                    if 'bit' in node_style:
                        bit = node_style['bit']
                        signal_name = f'({bit}){signal_name}'
                    gtkw.trace(signal_name, color=node_color,
                               datafmt=node_base, alias=display)

        walk(gtkw_dom, root_style)



""" key strategic example showing how to do multi-input fan-in into a
    multi-stage pipeline, then multi-output fanout, with an unary muxid
    and cancellation

    the multiplex ID from the fan-in is passed in to the pipeline, preserved,
    and used as a routing ID on the fanout.
"""

from random import randint
from math import log
from nmigen import Module, Signal, Cat, Value, Elaboratable, Const
from nmigen.compat.sim import run_simulation
from nmigen.cli import verilog, rtlil

from nmutil.multipipe import CombMultiOutPipeline, CombMuxOutPipe
from nmutil.multipipe import PriorityCombMuxInPipe
from nmutil.singlepipe import MaskCancellable, RecordObject, Object

from nmutil.test import StepLimiter
import unittest


class PassData(Object):
    def __init__(self):
        Object.__init__(self)
        self.muxid = Signal(2, reset_less=True)
        self.idx = Signal(8, reset_less=True)
        self.data = Signal(16, reset_less=True)
        self.operator = Signal(2, reset_less=True)
        self.routeid = Signal(2, reset_less=True)  # muxidname


class PassThroughStage:
    def __init__(self):
        self.o = self.ospec()

    def ispec(self):
        return PassData()

    def ospec(self):
        return self.ispec()  # same as ospec

    def _setup(self, m, i):
        comb = m.d.comb
        #comb += self.o.eq(i)

    def process(self, i):
        return i


class SplitRouteStage:
    def __init__(self):
        self.o = self.ospec()

    def ispec(self):
        return PassData()

    def ospec(self):
        return PassData()

    def setup(self, m, i):
        comb = m.d.comb
        comb += self.o.eq(i)
        with m.If(i.operator == Const(1, 2)):
            # selects 2nd output in CombMuxOutPipe
            comb += self.o.routeid.eq(1)
            comb += self.o.data.eq(i.data + 1)  # add 1 to say "we did it"
            comb += self.o.operator.eq(2)  # don't get into infinite loop
        with m.Else():
            # selects 2nd output in CombMuxOutPipe
            comb += self.o.routeid.eq(0)

    def process(self, i):
        return self.o


class DecisionPipe(MaskCancellable):
    def __init__(self, maskwid):
        stage = SplitRouteStage()
        MaskCancellable.__init__(self, stage, maskwid)


class RouteBackPipe(CombMuxOutPipe):
    """ routes data back to start of pipeline
    """

    def __init__(self):
        stage = PassThroughStage()
        CombMuxOutPipe.__init__(self, stage, n_len=2,
                                maskwid=4, muxidname="routeid",
                                routemask=True)


class MergeRoutePipe(PriorityCombMuxInPipe):
    """ merges data coming from end of pipe (with operator now == 1)
    """

    def __init__(self):
        stage = PassThroughStage()
        PriorityCombMuxInPipe.__init__(self, stage, p_len=2, maskwid=4,
                                       routemask=True)


class PassThroughPipe(MaskCancellable):
    def __init__(self, maskwid):
        MaskCancellable.__init__(self, PassThroughStage(), maskwid)


class InputTest:
    def __init__(self, dut, tlen):
        self.dut = dut
        self.di = {}
        self.do = {}
        self.sent = {}
        self.tlen = tlen
        for muxid in range(dut.num_rows):
            self.di[muxid] = {}
            self.do[muxid] = {}
            self.sent[muxid] = []
            for i in range(self.tlen):
                self.di[muxid][i] = randint(0, 255) + (muxid << 8)
                self.do[muxid][i] = self.di[muxid][i]

    def send(self, muxid):
        for i in range(self.tlen):
            op2 = self.di[muxid][i]
            rs = self.dut.p[muxid]
            yield rs.valid_i.eq(1)
            yield rs.data_i.data.eq(op2)
            yield rs.data_i.idx.eq(i)
            yield rs.data_i.muxid.eq(muxid)
            yield rs.data_i.operator.eq(1)
            yield rs.mask_i.eq(1)
            yield
            o_p_ready = yield rs.ready_o
            step_limiter = StepLimiter(10000)
            while not o_p_ready:
                step_limiter.step()
                yield
                o_p_ready = yield rs.ready_o

            print("send", muxid, i, hex(op2), op2)
            self.sent[muxid].append(i)

            yield rs.valid_i.eq(0)
            yield rs.mask_i.eq(0)
            # wait until it's received
            step_limiter = StepLimiter(10000)
            while i in self.do[muxid]:
                step_limiter.step()
                yield

            # wait random period of time before queueing another value
            for i in range(randint(0, 3)):
                yield

        yield rs.valid_i.eq(0)
        yield

        print("send ended", muxid)

        # wait random period of time before queueing another value
        # for i in range(randint(0, 3)):
        #    yield

        #send_range = randint(0, 3)
        # if send_range == 0:
        #    send = True
        # else:
        #    send = randint(0, send_range) != 0

    def rcv(self, muxid):
        rs = self.dut.p[muxid]
        for _ in StepLimiter(10000):

            # check cancellation
            if False and self.sent[muxid] and randint(0, 2) == 0:
                todel = self.sent[muxid].pop()
                print("to delete", muxid, self.sent[muxid], todel)
                if todel in self.do[muxid]:
                    del self.do[muxid][todel]
                    yield rs.stop_i.eq(1)
                print("left", muxid, self.do[muxid])
                if len(self.do[muxid]) == 0:
                    break

            #stall_range = randint(0, 3)
            # for j in range(randint(1,10)):
            #    stall = randint(0, stall_range) != 0
            #    yield self.dut.n[0].ready_i.eq(stall)
            #    yield

            n = self.dut.n[muxid]
            yield n.ready_i.eq(1)
            yield
            yield rs.stop_i.eq(0)  # resets cancel mask
            o_n_valid = yield n.valid_o
            i_n_ready = yield n.ready_i
            if not o_n_valid or not i_n_ready:
                continue

            out_muxid = yield n.data_o.muxid
            out_i = yield n.data_o.idx
            out_v = yield n.data_o.data

            print("recv", out_muxid, out_i, hex(out_v), hex(out_v))

            # see if this output has occurred already, delete it if it has
            assert muxid == out_muxid, \
                "out_muxid %d not correct %d" % (out_muxid, muxid)
            if out_i not in self.sent[muxid]:
                print("cancelled/recv", muxid, out_i)
                continue
            assert out_i in self.do[muxid], "out_i %d not in array %s" % \
                (out_i, repr(self.do[muxid]))
            assert self.do[muxid][out_i] + 1 == out_v  # check data
            del self.do[muxid][out_i]
            todel = self.sent[muxid].index(out_i)
            del self.sent[muxid][todel]

            # check if there's any more outputs
            if len(self.do[muxid]) == 0:
                break

        print("recv ended", muxid)


class TestPriorityMuxPipe(PriorityCombMuxInPipe):
    def __init__(self, num_rows):
        self.num_rows = num_rows
        stage = PassThroughStage()
        PriorityCombMuxInPipe.__init__(self, stage,
                                       p_len=self.num_rows, maskwid=1)


class TestMuxOutPipe(CombMuxOutPipe):
    def __init__(self, num_rows):
        self.num_rows = num_rows
        stage = PassThroughStage()
        CombMuxOutPipe.__init__(self, stage, n_len=self.num_rows,
                                maskwid=1)


class TestInOutPipe(Elaboratable):
    def __init__(self, num_rows=4):
        self.num_rows = nr = num_rows
        self.inpipe = TestPriorityMuxPipe(nr)  # fan-in (combinatorial)
        self.mergein = MergeRoutePipe()       # merge in feedback
        self.pipe1 = PassThroughPipe(nr)      # stage 1 (clock-sync)
        self.pipe2 = DecisionPipe(nr)         # stage 2 (clock-sync)
        # self.pipe3 = PassThroughPipe(nr)      # stage 3 (clock-sync)
        # self.pipe4 = PassThroughPipe(nr)      # stage 4 (clock-sync)
        self.splitback = RouteBackPipe()      # split back to mergein
        self.outpipe = TestMuxOutPipe(nr)     # fan-out (combinatorial)
        self.fifoback = PassThroughPipe(nr)   # temp route-back store

        self.p = self.inpipe.p  # kinda annoying,
        self.n = self.outpipe.n  # use pipe in/out as this class in/out
        self._ports = self.inpipe.ports() + self.outpipe.ports()

    def elaborate(self, platform):
        m = Module()
        m.submodules.inpipe = self.inpipe
        m.submodules.mergein = self.mergein
        m.submodules.pipe1 = self.pipe1
        m.submodules.pipe2 = self.pipe2
        #m.submodules.pipe3 = self.pipe3
        #m.submodules.pipe4 = self.pipe4
        m.submodules.splitback = self.splitback
        m.submodules.outpipe = self.outpipe
        m.submodules.fifoback = self.fifoback

        m.d.comb += self.inpipe.n.connect_to_next(self.mergein.p[0])
        m.d.comb += self.mergein.n.connect_to_next(self.pipe1.p)
        m.d.comb += self.pipe1.connect_to_next(self.pipe2)
        #m.d.comb += self.pipe2.connect_to_next(self.pipe3)
        #m.d.comb += self.pipe3.connect_to_next(self.pipe4)
        m.d.comb += self.pipe2.connect_to_next(self.splitback)
        m.d.comb += self.splitback.n[1].connect_to_next(self.fifoback.p)
        m.d.comb += self.fifoback.n.connect_to_next(self.mergein.p[1])
        m.d.comb += self.splitback.n[0].connect_to_next(self.outpipe.p)

        return m

    def ports(self):
        return self._ports


@unittest.skip("buggy -- fails due to exceeding step count limit")  # FIXME
def test1():
    dut = TestInOutPipe()
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_inoutmux_feedback_pipe.il", "w") as f:
        f.write(vl)

    tlen = 5

    test = InputTest(dut, tlen)
    run_simulation(dut, [test.rcv(0),  # test.rcv(1),
                         #test.rcv(3), test.rcv(2),
                         test.send(0),  # test.send(1),
                         #test.send(3), test.send(2),
                         ],
                   vcd_name="test_inoutmux_feedback_pipe.vcd")


if __name__ == '__main__':
    #from cProfile import Profile
    #p = Profile()
    # p.enable()
    test1()
    # p.disable()
    # p.print_stats()



""" concurrent unit from mitch alsup augmentations to 6600 scoreboard

    This work is funded through NLnet under Grant 2019-02-012

    License: LGPLv3+


    * data fans in
    * data goes through a pipeline
    * results fan back out.

    the output data format has to have a member "muxid", which is used
    as the array index on fan-out

    Associated bugreports:

    * https://bugs.libre-soc.org/show_bug.cgi?id=538
"""

from math import log
from nmigen import Module, Elaboratable, Signal
from nmigen.asserts import Assert
from nmigen.cli import main, verilog

from nmutil.singlepipe import PassThroughStage
from nmutil.multipipe import CombMuxOutPipe
from nmutil.multipipe import PriorityCombMuxInPipe


def num_bits(n):
    return int(log(n) / log(2))


class PipeContext:

    def __init__(self, pspec):
        """ creates a pipeline context.  currently: operator (op) and muxid

            opkls (within pspec) - the class to create that will be the
                                   "operator". instance must have an "eq"
                                   function.
        """
        self.id_wid = pspec.id_wid
        self.op_wid = pspec.op_wid
        self.muxid = Signal(self.id_wid, reset_less=True)   # RS multiplex ID
        opkls = pspec.opkls
        if opkls is None:
            self.op = Signal(self.op_wid, reset_less=True)
        else:
            self.op = opkls(pspec)

    def eq(self, i):
        ret = [self.muxid.eq(i.muxid)]
        ret.append(self.op.eq(i.op))
        # don't forget to update matches if you add fields later.
        return ret

    def matches(self, another):
        """
        Returns a list of Assert()s validating that this context
        matches the other context.
        """
        # I couldn't figure a clean way of overloading the == operator.
        return [
            Assert(self.muxid == another.muxid),
            Assert(self.op == another.op),
        ]

    def __iter__(self):
        yield self.muxid
        yield self.op

    def ports(self):
        if hasattr(self.op, "ports"):
            return [self.muxid] + self.op.ports()
        else:
            return list(self)


class InMuxPipe(PriorityCombMuxInPipe):
    def __init__(self, num_rows, iospecfn, maskwid=0):
        self.num_rows = num_rows
        stage = PassThroughStage(iospecfn)
        PriorityCombMuxInPipe.__init__(self, stage, p_len=self.num_rows,
                                       maskwid=maskwid)


class MuxOutPipe(CombMuxOutPipe):
    def __init__(self, num_rows, iospecfn, maskwid=0):
        self.num_rows = num_rows
        stage = PassThroughStage(iospecfn)
        CombMuxOutPipe.__init__(self, stage, n_len=self.num_rows,
                                maskwid=maskwid)


class ReservationStations(Elaboratable):
    """ Reservation-Station pipeline

        Input: num_rows - number of input and output Reservation Stations

        Requires: the addition of an "alu" object, from which ispec and ospec
        are taken, and inpipe and outpipe are connected to it

        * fan-in on inputs (an array of BaseData: a,b,mid)
        * ALU pipeline
        * fan-out on outputs (an array of FPPackData: z,mid)

        Fan-in and Fan-out are combinatorial.
    """
    def __init__(self, num_rows, maskwid=0, feedback_width=None):
        self.num_rows = nr = num_rows
        self.feedback_width = feedback_width
        self.inpipe = InMuxPipe(nr, self.i_specfn, maskwid)   # fan-in
        self.outpipe = MuxOutPipe(nr, self.o_specfn, maskwid) # fan-out

        self.p = self.inpipe.p  # kinda annoying,
        self.n = self.outpipe.n # use pipe in/out as this class in/out
        self._ports = self.inpipe.ports() + self.outpipe.ports()

    def elaborate(self, platform):
        m = Module()
        m.submodules.inpipe = self.inpipe
        m.submodules.alu = self.alu
        m.submodules.outpipe = self.outpipe

        m.d.comb += self.inpipe.n.connect_to_next(self.alu.p)
        m.d.comb += self.alu.connect_to_next(self.outpipe)

        if self.feedback_width is None:
            return m

        # connect all outputs above the feedback width back to their inputs
        # (hence, feedback).  pipeline stages are then expected to *modify*
        # the muxid (with care) in order to use the "upper numbered" RSes
        # for storing partially-completed results.  micro-coding, basically

        for i in range(self.feedback_width, self.num_rows):
            self.outpipe.n[i].connect_to_next(self.inpipe.p[i])

        return m

    def ports(self):
        return self._ports

    def i_specfn(self):
        return self.alu.ispec()

    def o_specfn(self):
        return self.alu.ospec()



""" Pipeline and BufferedHandshake examples
"""

from nmutil.nmoperator import eq
from nmutil.iocontrol import (PrevControl, NextControl)
from nmutil.singlepipe import (PrevControl, NextControl, ControlBase,
                        StageCls, Stage, StageChain,
                        BufferedHandshake, UnbufferedPipeline)

from nmigen import Signal, Module
from nmigen.cli import verilog, rtlil


class ExampleAddStage(StageCls):
    """ an example of how to use the buffered pipeline, as a class instance
    """

    def ispec(self):
        """ returns a tuple of input signals which will be the incoming data
        """
        return (Signal(16), Signal(16))

    def ospec(self):
        """ returns an output signal which will happen to contain the sum
            of the two inputs
        """
        return Signal(16)

    def process(self, i):
        """ process the input data (sums the values in the tuple) and returns it
        """
        return i[0] + i[1]


class ExampleBufPipeAdd(BufferedHandshake):
    """ an example of how to use the buffered pipeline, using a class instance
    """

    def __init__(self):
        addstage = ExampleAddStage()
        BufferedHandshake.__init__(self, addstage)


class ExampleStage(Stage):
    """ an example of how to use the buffered pipeline, in a static class
        fashion
    """

    def ispec():
        return Signal(16, name="example_input_signal")

    def ospec():
        return Signal(16, name="example_output_signal")

    def process(i):
        """ process the input data and returns it (adds 1)
        """
        return i + 1


class ExampleStageCls(StageCls):
    """ an example of how to use the buffered pipeline, in a static class
        fashion
    """

    def ispec(self):
        return Signal(16, name="example_input_signal")

    def ospec(self):
        return Signal(16, name="example_output_signal")

    def process(self, i):
        """ process the input data and returns it (adds 1)
        """
        return i + 1


class ExampleBufPipe(BufferedHandshake):
    """ an example of how to use the buffered pipeline.
    """

    def __init__(self):
        BufferedHandshake.__init__(self, ExampleStage)


class ExamplePipeline(UnbufferedPipeline):
    """ an example of how to use the unbuffered pipeline.
    """

    def __init__(self):
        UnbufferedPipeline.__init__(self, ExampleStage)


if __name__ == '__main__':
    dut = ExampleBufPipe()
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_bufpipe.il", "w") as f:
        f.write(vl)

    dut = ExamplePipeline()
    vl = rtlil.convert(dut, ports=dut.ports())
    with open("test_combpipe.il", "w") as f:
        f.write(vl)



"""Generation of GTKWave documents with nmutil.gtkw"""

from nmigen import Elaboratable, Signal, Module, Cat
from nmigen.back.pysim import Simulator
from nmigen.cli import rtlil
from math import log2

from vcd.gtkw import GTKWSave, GTKWColor
from nmutil.gtkw import write_gtkw


class Shifter(Elaboratable):
    """Simple sequential shifter

    * "Prev" port:

        * ``p_data_i``: value to be shifted

        * ``p_shift_i``: shift amount

        * ``op__sdir``: shift direction (0 = left, 1 = right)

        * ``p_valid_i`` and ``p_ready_o``: handshake

    * "Next" port:

        * ``n_data_o``: shifted value

        * ``n_valid_o`` and ``n_ready_i``: handshake
    """
    def __init__(self, width):
        self.width = width
        """data width"""
        self.p_data_i = Signal(width)
        self.p_shift_i = Signal(width)
        self.op__sdir = Signal()
        self.p_valid_i = Signal()
        self.p_ready_o = Signal()
        self.n_data_o = Signal(width)
        self.n_valid_o = Signal()
        self.n_ready_i = Signal()

    def elaborate(self, _):
        m = Module()

        # the control signals
        load = Signal()
        shift = Signal()
        direction = Signal()
        # the data flow
        shift_in = Signal(self.width)
        shift_left_by_1 = Signal(self.width)
        shift_right_by_1 = Signal(self.width)
        next_shift = Signal(self.width)
        # the register
        shift_reg = Signal(self.width, reset_less=True)
        # build the data flow
        m.d.comb += [
            # connect input and output
            shift_in.eq(self.p_data_i),
            self.n_data_o.eq(shift_reg),
            # generate shifted views of the register
            shift_left_by_1.eq(Cat(0, shift_reg[:-1])),
            shift_right_by_1.eq(Cat(shift_reg[1:], 0)),
        ]
        # choose the next value of the register according to the
        # control signals
        # default is no change
        m.d.comb += next_shift.eq(shift_reg)
        with m.If(load):
            m.d.comb += next_shift.eq(shift_in)
        with m.Elif(shift):
            with m.If(direction):
                m.d.comb += next_shift.eq(shift_right_by_1)
            with m.Else():
                m.d.comb += next_shift.eq(shift_left_by_1)

        # register the next value
        m.d.sync += shift_reg.eq(next_shift)

        # Shift counter
        shift_width = int(log2(self.width)) + 1
        next_count = Signal(shift_width)
        count = Signal(shift_width, reset_less=True)
        m.d.sync += count.eq(next_count)

        with m.FSM():
            with m.State("IDLE"):
                m.d.comb += [
                    # keep p.ready_o active on IDLE
                    self.p_ready_o.eq(1),
                    # keep loading the shift register and shift count
                    load.eq(1),
                    next_count.eq(self.p_shift_i),
                ]
                # capture the direction bit as well
                m.d.sync += direction.eq(self.op__sdir)
                with m.If(self.p_valid_i):
                    # Leave IDLE when data arrives
                    with m.If(next_count == 0):
                        # short-circuit for zero shift
                        m.next = "DONE"
                    with m.Else():
                        m.next = "SHIFT"
            with m.State("SHIFT"):
                m.d.comb += [
                    # keep shifting, while counter is not zero
                    shift.eq(1),
                    # decrement the shift counter
                    next_count.eq(count - 1),
                ]
                with m.If(next_count == 0):
                    # exit when shift counter goes to zero
                    m.next = "DONE"
            with m.State("DONE"):
                # keep n_valid_o active while the data is not accepted
                m.d.comb += self.n_valid_o.eq(1)
                with m.If(self.n_ready_i):
                    # go back to IDLE when the data is accepted
                    m.next = "IDLE"

        return m

    def __iter__(self):
        yield self.op__sdir
        yield self.p_data_i
        yield self.p_shift_i
        yield self.p_valid_i
        yield self.p_ready_o
        yield self.n_ready_i
        yield self.n_valid_o
        yield self.n_data_o

    def ports(self):
        return list(self)


def write_gtkw_direct():
    """Write a formatted GTKWave "save" file, using vcd.gtkw directly"""
    # hierarchy path, to prepend to signal names
    dut = "top.shf."
    # color styles
    style_input = GTKWColor.orange
    style_output = GTKWColor.yellow
    style_debug = GTKWColor.red
    with open("test_shifter_direct.gtkw", "wt") as gtkw_file:
        gtkw = GTKWSave(gtkw_file)
        gtkw.comment("Auto-generated by " + __file__)
        gtkw.dumpfile("test_shifter.vcd")
        # set a reasonable zoom level
        # also, move the marker to an interesting place
        gtkw.zoom_markers(-22.9, 10500000)
        gtkw.trace(dut + "clk")
        # place a comment in the signal names panel
        gtkw.blank("Shifter Demonstration")
        with gtkw.group("prev port"):
            gtkw.trace(dut + "op__sdir", color=style_input)
            # demonstrates using decimal base (default is hex)
            gtkw.trace(dut + "p_data_i[7:0]", color=style_input,
                       datafmt='dec')
            gtkw.trace(dut + "p_shift_i[7:0]", color=style_input,
                       datafmt='dec')
            gtkw.trace(dut + "p_valid_i", color=style_input)
            gtkw.trace(dut + "p_ready_o", color=style_output)
        with gtkw.group("debug"):
            gtkw.blank("Some debug statements")
            # change the displayed name in the panel
            gtkw.trace("top.zero", alias='zero delay shift',
                       color=style_debug)
            gtkw.trace("top.interesting", color=style_debug)
            gtkw.trace("top.test_case", alias="test case", color=style_debug)
            gtkw.trace("top.msg", color=style_debug)
        with gtkw.group("internal"):
            gtkw.trace(dut + "fsm_state")
            gtkw.trace(dut + "count[3:0]")
            gtkw.trace(dut + "shift_reg[7:0]", datafmt='dec')
        with gtkw.group("next port"):
            gtkw.trace(dut + "n_data_o[7:0]", color=style_output,
                       datafmt='dec')
            gtkw.trace(dut + "n_valid_o", color=style_output)
            gtkw.trace(dut + "n_ready_i", color=style_input)


def test_shifter():
    """Simulate the Shifter to generate some traces,
     as well as the GTKWave documents"""
    m = Module()
    m.submodules.shf = dut = Shifter(8)
    print("Shifter port names:")
    for port in dut:
        print("-", port.name)
    # generate RTLIL
    # try "proc; show" in yosys to check the data path
    il = rtlil.convert(dut, ports=dut.ports())
    with open("test_shifter.il", "w") as f:
        f.write(il)

    # write the GTKWave project file, directly
    write_gtkw_direct()

    # Describe a GTKWave document

    # Style for signals, classes and groups
    gtkwave_style = {
        # Root selector. Gives default attributes for every signal.
        '': {'base': 'dec'},
        # color the traces, according to class
        # class names are not hardcoded, they are just strings
        'in': {'color': 'orange'},
        'out': {'color': 'yellow'},
        # signals in the debug group have a common color and module path
        'debug': {'module': 'top', 'color': 'red'},
        # display a different string replacing the signal name
        'test_case': {'display': 'test case'},
    }

    # DOM style description for the trace pane
    gtkwave_desc = [
        # simple signal, without a class
        # even so, it inherits the top-level root attributes
        'clk',
        # comment
        {'comment': 'Shifter Demonstration'},
        # collapsible signal group
        ('prev port', [
            # attach a class style for each signal
            ('op__sdir', 'in'),
            ('p_data_i[7:0]', 'in'),
            ('p_shift_i[7:0]', 'in'),
            ('p_valid_i', 'in'),
            ('p_ready_o', 'out'),
        ]),
        # Signals in a signal group inherit the group attributes.
        # In this case, a different module path and color.
        ('debug', [
            {'comment': 'Some debug statements'},
            # inline attributes, instead of a class name
            ('zero', {'display': 'zero delay shift'}),
            'interesting',
            'test_case',
            'msg',
        ]),
        ('internal', [
            'fsm_state',
            'count[3:0]',
            'shift_reg[7:0]',
        ]),
        ('next port', [
            ('n_data_o[7:0]', 'out'),
            ('n_valid_o', 'out'),
            ('n_ready_i', 'in'),
        ]),
    ]

    write_gtkw("test_shifter.gtkw", "test_shifter.vcd",
               gtkwave_desc,  gtkwave_style,
               module="top.shf", loc=__file__, marker=10500000)

    sim = Simulator(m)
    sim.add_clock(1e-6)

    # demonstrates adding extra debug signal traces
    # they end up in the top module
    #
    zero = Signal()  # mark an interesting place
    #
    # demonstrates string traces
    #
    # display a message when the signal is high
    # the low level is just an horizontal line
    interesting = Signal(decoder=lambda v: 'interesting!' if v else '')
    # choose between alternate strings based on numerical value
    test_cases = ['', '13>>2', '3<<4', '21<<0']
    test_case = Signal(8, decoder=lambda v: test_cases[v])
    # hack to display arbitrary strings, like debug statements
    msg = Signal(decoder=lambda _: msg.str)
    msg.str = ''

    def send(data, shift, direction):
        # present input data and assert valid_i
        yield dut.p_data_i.eq(data)
        yield dut.p_shift_i.eq(shift)
        yield dut.op__sdir.eq(direction)
        yield dut.p_valid_i.eq(1)
        yield
        # wait for p.ready_o to be asserted
        while not (yield dut.p_ready_o):
            yield
        # show current operation operation
        if direction:
            msg.str = f'{data}>>{shift}'
        else:
            msg.str = f'{data}<<{shift}'
        # force dump of the above message by toggling the
        # underlying signal
        yield msg.eq(0)
        yield msg.eq(1)
        # clear input data and negate p.valid_i
        yield dut.p_valid_i.eq(0)
        yield dut.p_data_i.eq(0)
        yield dut.p_shift_i.eq(0)
        yield dut.op__sdir.eq(0)

    def receive(expected):
        # signal readiness to receive data
        yield dut.n_ready_i.eq(1)
        yield
        # wait for n.valid_o to be asserted
        while not (yield dut.n_valid_o):
            yield
        # read result
        result = yield dut.n_data_o
        # negate n.ready_i
        yield dut.n_ready_i.eq(0)
        # check result
        assert result == expected
        # finish displaying the current operation
        msg.str = ''
        yield msg.eq(0)
        yield msg.eq(1)

    def producer():
        # 13 >> 2
        yield from send(13, 2, 1)
        # 3 << 4
        yield from send(3, 4, 0)
        # 21 << 0
        # use a debug signal to mark an interesting operation
        # in this case, it is a shift by zero
        yield interesting.eq(1)
        yield from send(21, 0, 0)
        yield interesting.eq(0)

    def consumer():
        # the consumer is not in step with the producer, but the
        # order of the results are preserved
        # 13 >> 2 = 3
        yield test_case.eq(1)
        yield from receive(3)
        # 3 << 4 = 48
        yield test_case.eq(2)
        yield from receive(48)
        # 21 << 0 = 21
        yield test_case.eq(3)
        # you can look for the rising edge of this signal to quickly
        # locate this point in the traces
        yield zero.eq(1)
        yield from receive(21)
        yield zero.eq(0)
        yield test_case.eq(0)

    sim.add_sync_process(producer)
    sim.add_sync_process(consumer)
    sim_writer = sim.write_vcd(
        "test_shifter.vcd",
        # include additional signals in the trace dump
        traces=[zero, interesting, test_case, msg],
    )
    with sim_writer:
        sim.run()


if __name__ == "__main__":
    test_shifter()
